/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: iptools.cc                                   */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "iptools.h"

// **************************************
// ************ IP Netmask **************
// **************************************

/* Warning: using this code assumes that your compiler is enabled to
   handle circular values (e.g. when an unsigned char variable c is
   set to 255, c+1 returns 0);
   moreover, partial expression evaluation (e.g. evaluation of just A
   in (A||B) when A is true) should be used for efficiency reasons */

// Default constructor
IPNetmask::IPNetmask() {
   bytes[0] = bytes[1] = bytes[2] = bytes[3] = 255;
}

// Destructor
IPNetmask::~IPNetmask() {
}

// Copy constructor
IPNetmask::IPNetmask (const IPNetmask& source) {
   bytes[0] = source.bytes[0];
   bytes[1] = source.bytes[1];
   bytes[2] = source.bytes[2];
   bytes[3] = source.bytes[3];
}

/* Builds a netmask from its four bytes; illegal netmasks are replaced
   with 255.255.255.255 */
IPNetmask::IPNetmask(unsigned char byte0, unsigned char byte1,
                     unsigned char byte2, unsigned char byte3) {
   setNetmask(byte0, byte1, byte2, byte3);
}

/* Builds a netmask from a value that represents the number of
   bits assigned to the network */
IPNetmask::IPNetmask(char netBits) {
   setNetmask(netBits);
}

// Assignment operator overload
IPNetmask& IPNetmask::operator=(const IPNetmask& source) {
   bytes[0] = source.bytes[0];
   bytes[1] = source.bytes[1];
   bytes[2] = source.bytes[2];
   bytes[3] = source.bytes[3];
   return (*this);
}

// Returns true if the two objects represent the same netmask
bool IPNetmask::operator==(const IPNetmask& otherNetmask) {
   return (bytes[0] == otherNetmask.bytes[0] &&
           bytes[1] == otherNetmask.bytes[1] &&
           bytes[2] == otherNetmask.bytes[2] &&
           bytes[3] == otherNetmask.bytes[3]);
}

// Inequality operator
bool IPNetmask::operator!=(const IPNetmask& otherNetmask) {
   return (!(*this==otherNetmask));
}

// Returns true if the lvalue netmask has more `1's than the other one
bool IPNetmask::operator>(const IPNetmask& otherNetmask) {
   for (unsigned char i=0; i<=3; i++) {
      if (bytes[i] > otherNetmask.bytes[i]) {
         return true;
      }
   }
   return false;
}

// Greater than or equal to operator overload
bool IPNetmask::operator>=(const IPNetmask& otherNetmask) {
   return ((*this)>otherNetmask || (*this)==otherNetmask);
}

// Returns true if the lvalue netmask has less `1's than the other one
bool IPNetmask::operator<(const IPNetmask& otherNetmask) {
   for (unsigned char i=0; i<=3; i++) {
      if (bytes[i] < otherNetmask.bytes[i]) {
         return true;
      }
   }
   return false;
}

// Less than or equal to operator overload
bool IPNetmask::operator<=(const IPNetmask& otherNetmask) {
   return ((*this)<otherNetmask || (*this)==otherNetmask);
}

/* Changes the netmask by setting its four bytes; illegal netmasks are
   replaced with 255.255.255.255 */
void IPNetmask::setNetmask(unsigned char byte0, unsigned char byte1,
                           unsigned char byte2, unsigned char byte3) {
   unsigned char i, bit, curByte;
   bool wrongByte, hasZeros, foundOne;
   bytes[0] = byte0;
   bytes[1] = byte1;
   bytes[2] = byte2;
   bytes[3] = byte3;
   // Check if netmask bits are correct
   wrongByte = hasZeros = false;
   i = 0;
   while (!wrongByte && i<=3) {
      curByte = bytes[i];
      if (hasZeros && (curByte != 0)) {
         // A previous byte had zeros: this one cannot be non-zero
         wrongByte = true;
      }
      else {
         if (curByte != 0xFF) {
            hasZeros = true;
         }
         if (curByte != 0 && hasZeros) {
            // This is a non-zero and non-FF byte: check it
            /* A negated netmask byte plus 1 must result
               in a mask having at most one bit set to 1 */
            curByte = ~bytes[i] + 1;
            bit = 0;
            do {
               foundOne = (curByte & 0x80) != 0;      // Check first bit
               curByte <<= 1;
               bit++;
            } while (bit <= 7 && !foundOne);
            // Are there other `1's beyond the found one?
            wrongByte = (curByte != 0);
         }
      }
      i++;
   }

   if (wrongByte) {
      // This mask contains an illegal byte: replace it with 255.255.255.255
      bytes[0] = 255;
      bytes[1] = 255;
      bytes[2] = 255;
      bytes[3] = 255;
   }
}

/* Changes the netmask on the basis of its `brief' value (i.e., the
   number of bytes that define the network) */
void IPNetmask::setNetmask(char netBits) {
   if (netBits < 0 || netBits > 32) {
      // Assume invalid netmask stands for "single address"
      netBits = 32;
   }
   for (unsigned char i=0; i<=3; i++) {
      // The following expression computes a sequence of 1's followed by 0's
      if (netBits>=8) {
         // 1<<8 = 2^8; thus, (1<<8)-1 = 2^8-1 = 1111 1111
         bytes[i] = (1 << 8)-1;
         netBits -= 8;
      }
      else {
         /* Example that justifies the following expression:
                       netmask     = 3 (expected bitmask: 1110 0000)
                    (8-netmask)    = 5
                 1<<(8-netmask)    = 2^5 (0010 0000)
                 1<<(8-netmask)-1  =      0001 1111
               ~(1<<(8-netmask)-1) =      1110 0000 (as expected) */
         bytes[i] = ~ ((1 << (8 - netBits)) - 1);
         netBits = 0;
      }
   }
}

/* Returns the four bytes of the netmask; warning: subsequent
   callings to this function always return the same address,
   so avoid storing it for later use because the bytes it points
   to can be altered by a later call to this method */
unsigned char *IPNetmask::getNetmask() {
   retBytes[0] = bytes[0];
   retBytes[1] = bytes[1];
   retBytes[2] = bytes[2];
   retBytes[3] = bytes[3];
   return retBytes;
}

/* Returns the netmask in the `brief' format (i.e. as the number of
   bits that form the network) */
char IPNetmask::brief() {
   char netmask;
   unsigned char bit, i;
   bool goOn;
   goOn = true;
   netmask = 0;
   i = 0;
   while (goOn && i<=3) {
      bit = 1;
      while (goOn && bit<=8) {
         // (1 << (8-bit)) is a mask that has only the `bit'th bit to 1
         goOn = bytes[i] & (1 << (8-bit));
         netmask += (goOn ? 1 : 0);
         bit++;
      }
      i++;
   }
   return netmask;
}

// ************ IP Netmask **************






// **************************************
// ************ IP Address **************
// **************************************

/* Warning: using this code assumes that your compiler considers the range
   of allowed values for each data type as circular (e.g. if c is a char,
   when c=255 we have: c+1=0 */

// Default constructor
IPAddress::IPAddress() {
   addrByte[0] = 0;
   addrByte[1] = 0;
   addrByte[2] = 0;
   addrByte[3] = 0;
}

// Destructor
IPAddress::~IPAddress() {
}

// Another constructor
IPAddress::IPAddress(unsigned char byte0, unsigned char byte1,
                     unsigned char byte2, unsigned char byte3) {
   addrByte[0] = byte0;
   addrByte[1] = byte1;
   addrByte[2] = byte2;
   addrByte[3] = byte3;
}

// Copy constructor
IPAddress::IPAddress(const IPAddress& source) {
   addrByte[0] = source.addrByte[0];
   addrByte[1] = source.addrByte[1];
   addrByte[2] = source.addrByte[2];
   addrByte[3] = source.addrByte[3];
}

// Assignment operator overload
IPAddress& IPAddress::operator=(const IPAddress& source) {
   addrByte[0] = source.addrByte[0];
   addrByte[1] = source.addrByte[1];
   addrByte[2] = source.addrByte[2];
   addrByte[3] = source.addrByte[3];
   return (*this);
}

// Equality test overload
bool IPAddress::operator==(const IPAddress& comp) {
   return (addrByte[0]==comp.addrByte[0] &&
           addrByte[1]==comp.addrByte[1] &&
           addrByte[2]==comp.addrByte[2] &&
           addrByte[3]==comp.addrByte[3]);
}

// Inequality test overload
bool IPAddress::operator!=(const IPAddress& comp) {
   return (!(*this==comp));
}

/* Greater than operator overload; ordering of IP addresses
   is based on the value of their bytes (the leftmost
   being the most significant) */
bool IPAddress::operator>(const IPAddress& comp) {
   for (unsigned char i=0; i<=3; i++) {
      if (addrByte[i]!=comp.addrByte[i]) {
         return (addrByte[i]>comp.addrByte[i]);
      }
   }
   // The two objects represent the same address
   return false;
}

// Greater than or equal to operator overload
bool IPAddress::operator>=(const IPAddress& comp) {
   return ((*this)>comp || (*this)==comp);
}

// Less than operator overload
bool IPAddress::operator<(const IPAddress& comp) {
   for (unsigned char i=0; i<=3; i++) {
      if (addrByte[i]!=comp.addrByte[i]) {
         return (addrByte[i]<comp.addrByte[i]);
      }
   }
   // The two objects represent the same address
   return false;
}

// Less than or equal to operator overload
bool IPAddress::operator<=(const IPAddress& comp) {
   return ((*this)<comp || (*this)==comp);
}

/* Plus sign overload; adding an integer value n to an address results
   in the n-th address that follows the given one in the IP address space */
IPAddress IPAddress::operator+(const unsigned long long int& span) {
   IPAddress newAddress;
   unsigned long int thisAddress;
   // Linearize address space
   thisAddress =  addrByte[0]*256*256*256 +
                  addrByte[1]*256*256 + addrByte[2]*256 + addrByte[3];
   thisAddress += span;
   for (int i=3; i>=0; i--) {
      newAddress.addrByte[i] = thisAddress % 256;
      thisAddress = (unsigned long int) ((float) thisAddress/(float) 256);
   }
   return newAddress;
}

/* Increment assignment operator overload; works the same as a=a+n,
   where a is an IPAddress and n a long int */
IPAddress& IPAddress::operator+=(const unsigned long long int& span) {
   (*this) = (*this) + span;
   return (*this);
}

/* Minus sign overload; the difference between two IP addresses is the
   number of addresses in the IP space that fall between the two given
   ones (including themselves); i.e., the minimum value that this
   operator returns is 1 (when applying it between an address and itself);
   returned result is positive, whichever the ordering of the two addresses is */
unsigned long long int IPAddress::operator-(const IPAddress& addr) {
   unsigned long int thisAddress, otherAddress;
   // Linearize address space
   thisAddress =  addrByte[0]*256*256*256 +
                  addrByte[1]*256*256 + addrByte[2]*256 + addrByte[3];
   otherAddress = addr.addrByte[0]*256*256*256 +
                  addr.addrByte[1]*256*256 +
                  addr.addrByte[2]*256 +
                  addr.addrByte[3];
   // Check address ordering, and return a positive result
   if (thisAddress>otherAddress) {
      return ((unsigned long long int) thisAddress - (unsigned long long int) otherAddress + 1);
   }
   else {
      return ((unsigned long long int) otherAddress - (unsigned long long int) thisAddress + 1);
   }
}

/* Minus sign overload; subtracting an integer value n from an address
   results in the n-th address that precedes the given one */
IPAddress IPAddress::operator-(const unsigned long long int& span) {
   IPAddress newAddress;
   unsigned long int thisAddress;
   // Linearize address space
   thisAddress =  addrByte[0]*256*256*256 +
                  addrByte[1]*256*256 + addrByte[2]*256 + addrByte[3];
   thisAddress -= span;
   for (int i=3; i>=0; i--) {
      newAddress.addrByte[i] = thisAddress % 256;
      thisAddress = (unsigned long int) ((float) thisAddress/(float) 256);
   }
   return newAddress;
}

/* Decrement assignment operator overload; works the same as a=a-n,
   where a is an IPAddress and n a long int */
IPAddress& IPAddress::operator-=(const unsigned long long int& span) {
   (*this) = (*this) - span;
   return (*this);
}

/* Returns the address class:
   'A' for A class (/8)
   'B' for B class (/16)
   'C' for C class (/24)
   'D' for D class (multicast)
   'E' for E class (future use)
   'U' for unknown class */
char IPAddress::ipClass() {
   char c = 'U';
   if (addrByte[0]>=1 && addrByte[0]<=127) { c = 'A'; }
   if (addrByte[0]>=128 && addrByte[0]<=191) { c = 'B'; }
   if (addrByte[0]>=192 && addrByte[0]<=223) { c = 'C'; }
   if (addrByte[0]>=224 && addrByte[0]<=239) { c = 'D'; }
   if (addrByte[0]>=240 && addrByte[0]<=247) { c = 'E'; }
   return c;
}

/* Computes the bitwise "and" of the current address with a mask;
   does not alter current address */
IPAddress IPAddress::mask(const IPNetmask& netmask) {
   IPAddress newAddress;
   /* Using `const' qualifier prevents from directly invoking
      methods over netmask */
   IPNetmask workNetmask = netmask;
   unsigned char *maskBytes;
   maskBytes = workNetmask.getNetmask();
   newAddress.addrByte[0] = addrByte[0] & maskBytes[0];
   newAddress.addrByte[1] = addrByte[1] & maskBytes[1];
   newAddress.addrByte[2] = addrByte[2] & maskBytes[2];
   newAddress.addrByte[3] = addrByte[3] & maskBytes[3];
   return newAddress;
}

/* Returns the four address bytes; warning: subsequent
   callings to this function always return the same address,
   so avoid storing it for later use because the bytes it
   points to can be altered by a later call to this method */
unsigned char* IPAddress::getAddress() {
   retAddrByte[0] = addrByte[0];
   retAddrByte[1] = addrByte[1];
   retAddrByte[2] = addrByte[2];
   retAddrByte[3] = addrByte[3];
   return retAddrByte;
}

// Sets the four address bytes
void IPAddress::setAddress(unsigned char byte0, unsigned char byte1,
                           unsigned char byte2, unsigned char byte3) {
   addrByte[0] = byte0;
   addrByte[1] = byte1;
   addrByte[2] = byte2;
   addrByte[3] = byte3;
}

// ************ IP Address **************






// ***************************************
// ************** IP Prefix **************
// ***************************************

// Default constructor
IPPrefix::IPPrefix() {
   addr = IPAddress (0, 0, 0, 0);
   mask = IPNetmask (0, 0, 0, 0);
}

// Destructor
IPPrefix::~IPPrefix() {
}

/* Builds a prefix using only an IP address; netmask is
   based upon address class */
IPPrefix::IPPrefix(const IPAddress& newAddress) {
   // Cannot work directly on `addr' because of const qualifier
   IPAddress a = newAddress;
   addr = a;
   switch (a.ipClass()) {
      case 'A':   mask = IPNetmask(8);
                  break;
      case 'B':   mask = IPNetmask(16);
                  break;
      case 'C':   mask = IPNetmask(24);
                  break;
      /* In case of a special address, assume it is a single
         address */
      default:    mask = IPNetmask(255, 255, 255, 255);
                  break;
   }
}

/* Typical definition of a prefix: address and netmask; a check
   is performed to ensure address matches the beginning
   of the address block implied by the netmask */
IPPrefix::IPPrefix(const IPAddress& newAddress, const IPNetmask& newNetmask) {
   /* Cannot do the following operation in a single step, because const
      qualifiers prevents from invoking methods on newAddress */
   addr = newAddress;
   addr = addr.mask(newNetmask);
   mask = newNetmask;
}

// Assignment operator overload
IPPrefix& IPPrefix::operator=(const IPPrefix& newPrefix) {
   addr = newPrefix.addr;
   mask = newPrefix.mask;
   return (*this);
}

/* Returns true if the two objects represent the same prefix
   (i.e., they have the same address and netmask) */
bool IPPrefix::operator==(const IPPrefix& otherPrefix) {
   return (addr == otherPrefix.addr &&
           mask == otherPrefix.mask);
}

// Inequality operator overload
bool IPPrefix::operator!=(const IPPrefix& otherPrefix) {
   return (!(*this==otherPrefix));
}

/* Returns true only when current object represents a prefix
   that is more specific than a given one */
bool IPPrefix::isMoreSpecific(const IPPrefix& otherPrefix) {
   // Cannot directly work on otherPrefix because of `const' qualifier
   IPPrefix pfx = otherPrefix;
   return (mask > pfx.mask &&
           addr.mask(pfx.mask) == pfx.addr);
}

// The same as `isMoreSpecific', but in reverse order
bool IPPrefix::isLessSpecific(const IPPrefix& otherPrefix) {
   // Cannot directly work on otherPrefix because of `const' qualifier
   IPPrefix pfx = otherPrefix;
   return (mask < pfx.mask &&
           pfx.addr.mask(mask) == addr);
}

// Returns the address field of the prefix
IPAddress IPPrefix::address() {
   return addr;
}

// Returns the netmask of the prefix
IPNetmask IPPrefix::netmask() {
   return mask;
}

// Returns the number of addresses in the prefix
unsigned long long int IPPrefix::addresses() {
   return (1ULL << (32 - mask.brief()));
}

// Returns a list of the addresses that the prefix summarizes
List<IPAddress>& IPPrefix::listAddresses() {
   IPAddress i;
   bool stop;
   const IPAddress lastAddress = addr + (addresses() - 1);
   // Initialize an internal empty list
   addrRetList = List<IPAddress>();
   i = addr;
   stop = false;
   while (!stop && i<=lastAddress) {
      /* Must be careful when handling circular values
         (like IP addresses are) */
      addrRetList.enqueue(i);
      if (i == IPAddress(255, 255, 255, 255)) {
         stop = true;
      }
      else {
         i+=1;
      }
   }
   return (addrRetList);
}

// ************** IP Prefix **************






// **************************************
// ************  IP Subspace  ***********
// **************************************

// Default constructor; builds an empty set
IPSubSpace::IPSubSpace() {
   addrCount = 0;
}

// Destructor
IPSubSpace::~IPSubSpace() {
}

// Copy constructor
IPSubSpace::IPSubSpace(const IPSubSpace& source) {
   addrSet = source.addrSet;
   addrCount = source.addrCount;
}

/* Another constructor; initializes the set with a given prefix
   (address, netmask) */
IPSubSpace::IPSubSpace(const IPPrefix& prefix) {
   IPInterval newInterval;
   // Cannot directly work on `prefix' because of const qualifier
   IPPrefix workPfx = prefix;
   /* IPPrefix class already ensures that provided address
      matches first address of CIDR block */
   newInterval.begin = workPfx.address();
   newInterval.end = newInterval.begin + (workPfx.addresses() - 1);
   addrSet.addItem(newInterval, newInterval.begin);
   addrCount = workPfx.addresses();
}

/* Like the previous constructor, but without a netmask; inserts a
   single address in the set */
IPSubSpace::IPSubSpace(const IPAddress& address) {
   IPInterval newInterval;
   /* Working with a single address implies that no care should
      be taken about CIDR aspects */
   newInterval.begin = newInterval.end = address;
   addrSet.addItem(newInterval, newInterval.begin);
   addrCount = 1;
}

// Assignment operator overload
IPSubSpace& IPSubSpace::operator=(const IPSubSpace& source) {
   addrSet = source.addrSet;
   addrCount = source.addrCount;
   return (*this);
}

// Tests whether the two subspaces hold the same addresses
bool IPSubSpace::operator==(const IPSubSpace& otherSubSpace) {
   List<IPInterval> sourceIntervals, destIntervals;
   IPInterval sourceInt, destInt;
   BST<IPInterval, IPAddress> otherAddrSet;
   /* Perform a preliminary check to see if the subspaces can
      hold the same addresses */
   if (addrCount != otherSubSpace.addrCount) {
      return false;
   }
   else {
      /* Cannot invoke methods on otherSubSpace.addrSet because of
         `const' qualifier */
      otherAddrSet = otherSubSpace.addrSet;
      sourceIntervals = addrSet.listItems();
      destIntervals = otherAddrSet.listItems();
      while (!sourceIntervals.isEmpty() && !destIntervals.isEmpty()) {
         /* The two address sets are ordered based on IP addresses,
            so IP intervals should be encountered in the same order too */
         sourceInt = sourceIntervals.pop();
         destInt = destIntervals.pop();
         if ((sourceInt.begin != destInt.begin) ||
            (sourceInt.end != destInt.end)) {
            return false;
         }
      }
      if (!sourceIntervals.isEmpty() || !destIntervals.isEmpty()) {
         /* One of the two subspaces holds more intervals than
            the other */
         return false;
      }
   }
   return true;
}

// Inequality operator overload
bool IPSubSpace::operator!=(const IPSubSpace& otherSubSpace) {
   /* First of all, perform a quick test to see whether the
      two subspaces are different */
   return ((addrCount != otherSubSpace.addrCount) ||
           !((*this)==otherSubSpace));
}

/* Returns true if current subspace includes a given one
   (strict inclusion) */
bool IPSubSpace::includes(const IPSubSpace& otherSubSpace) {
   /* First condition is checked to avoid single intervals
      check in some situations */
   return ((addrCount > otherSubSpace.addrCount) &&
           (intersect(otherSubSpace)==otherSubSpace));
}

/* Returns true if current subspace is included in a
   given one (strict inclusion) */
bool IPSubSpace::isIncluded(const IPSubSpace& otherSubSpace) {
   /* First condition is checked to avoid single intervals
      check in some situations */
   return ((addrCount < otherSubSpace.addrCount) &&
           (intersect(otherSubSpace)==(*this)));
}

/* Returns true if the current object represents the
   empty subspace (i.e. no address) */
bool IPSubSpace::isEmpty() {
   return (addrSet.isEmpty());
}

// Computes the intersection of two address sets
IPSubSpace IPSubSpace::intersect(const IPSubSpace& otherSubSpace) {
   bool doIntersection;
   List<IPInterval> sourceIntervals, destIntervals;
   /* Working with a `const' parameter implies that no method can be
      invoked on its members */
   BST<IPInterval, IPAddress> otherAddrSet = otherSubSpace.addrSet;
   IPInterval sourceInt, destInt, newInterval;
   IPSubSpace retSpace;
   sourceIntervals = otherAddrSet.listItems();
   destIntervals = addrSet.listItems();
   if (sourceIntervals.isEmpty() || destIntervals.isEmpty()) {
      // There is nothing to intersect with
      return(retSpace);
   }
   // Intersection works like a merge, between couples of IP addresses intervals
   while (!sourceIntervals.isEmpty() || !destIntervals.isEmpty()) {
      if (!sourceIntervals.isEmpty()) {
         sourceInt = sourceIntervals.pop();
      }
      if (!destIntervals.isEmpty()) {
         destInt = destIntervals.pop();
      }
      doIntersection = true;
      if (sourceInt.end < destInt.begin) {
         // Must go forward on sourceIntervals list
         while (sourceInt.end < destInt.begin && !sourceIntervals.isEmpty()) {
            sourceInt = sourceIntervals.pop();
         }
         if (sourceIntervals.isEmpty()) {
            /* The search for an element to merge with has overcome the end of
               a list: there is nothing to intersect with */
            doIntersection = false;
         }
      }
      else {
         if (destInt.end < sourceInt.begin) {
            // Must go forward on destIntervals list
            while (destInt.end < sourceInt.begin && !destIntervals.isEmpty()) {
               destInt = destIntervals.pop();
            }
            if (destIntervals.isEmpty()) {
               /* The search for an element to merge with has overcome the end of
                  a list: there is nothing to intersect with */
               doIntersection = false;
            }
         }
      }
      if (doIntersection) {
         newInterval.begin = (sourceInt.begin > destInt.begin ? sourceInt.begin : destInt.begin);
         newInterval.end = (sourceInt.end < destInt.end ? sourceInt.end : destInt.end);
         retSpace.addrSet.addItem(newInterval, newInterval.begin);
         retSpace.addrCount += newInterval.end - newInterval.begin;
      }
   }
   return (retSpace);
}

// Alias for `intersect'
IPSubSpace& IPSubSpace::operator*(const IPSubSpace& otherSubSet) {
   (*this) = intersect(otherSubSet);
   return (*this);
}

// *= operator overload
IPSubSpace& IPSubSpace::operator*=(const IPSubSpace& otherSubSpace) {
   (*this) = (*this) * otherSubSpace;
   return (*this);
}

// Computes the union of two address sets
IPSubSpace IPSubSpace::join(const IPSubSpace& otherSubSpace) {
   bool merged, savedSource, savedDest;
   List<IPInterval> sourceIntervals, destIntervals;
   IPSubSpace retSpace;
   /* Working with a `const' parameter implies that no method can be
      invoked on its members */
   BST<IPInterval, IPAddress> otherAddrSet = otherSubSpace.addrSet;
   IPInterval sourceInt, destInt, newInterval;
   sourceIntervals = otherAddrSet.listItems();
   destIntervals = addrSet.listItems();
   savedSource = true;
   savedDest = true;
   merged = false;
   // If one of the subspaces is empty, the result is the other one
   if (sourceIntervals.isEmpty()) {
      retSpace.addrSet = addrSet;
      retSpace.addrCount = addrCount;
      return (retSpace);
   }
   if (destIntervals.isEmpty()) {
      retSpace.addrSet = otherAddrSet;
      retSpace.addrCount = otherSubSpace.addrCount;
      return (retSpace);
   }
   // Union works like a merge, between couples of IP addresses intervals
   while (!sourceIntervals.isEmpty() || !destIntervals.isEmpty()) {
      if (!sourceIntervals.isEmpty()) {
         sourceInt = sourceIntervals.pop();
         savedSource = false;
      }
      if (!destIntervals.isEmpty()) {
         destInt = destIntervals.pop();
         savedDest = false;
      }
      /* If one of the following conditions holds, intervals are disjoint
         and not consecutive: try to reach a couple of either overlapped
         or consecutive intervals; since IP addresses are circular values,
         care must be taken when dealing with minus sign */
      if ((destInt.begin > IPAddress()) && (sourceInt.end < destInt.begin - 1)) {
         /* Must go forward on sourceIntervals list; while doing that,
            insert all the encountered intervals in the set */
         if (!merged) {
            retSpace.addrSet.addItem(sourceInt, sourceInt.begin);
            retSpace.addrCount += sourceInt.end - sourceInt.begin;
         }
         savedSource = true;
         if (!sourceIntervals.isEmpty()) {
            sourceInt = sourceIntervals.pop();
            savedSource = false;
         }
         while ((destInt.begin > IPAddress()) && (sourceInt.end < destInt.begin - 1) &&
                !sourceIntervals.isEmpty()) {
            retSpace.addrSet.addItem(sourceInt, sourceInt.begin);
            retSpace.addrCount += sourceInt.end - sourceInt.begin;
            sourceInt = sourceIntervals.pop();
            savedSource = false;
         }
      }
      else {
         if ((sourceInt.begin > IPAddress()) && (destInt.end < sourceInt.begin - 1)) {
            /* Must go forward on destIntervals list; while doing that,
               insert all the encountered intervals in the set */
            if (!merged) {
               retSpace.addrSet.addItem(destInt, destInt.begin);
               retSpace.addrCount += destInt.end - destInt.begin;
            }
            savedDest = true;
            if (!destIntervals.isEmpty()) {
               destInt = destIntervals.pop();
               savedDest = false;
            }
            while ((sourceInt.begin > IPAddress()) && (destInt.end < sourceInt.begin - 1) &&
                   !destIntervals.isEmpty()) {
               retSpace.addrSet.addItem(destInt, destInt.begin);
               retSpace.addrCount += destInt.end - destInt.begin;
               destInt = destIntervals.pop();
               savedDest = false;
            }
         }
      }
      merged = false;
      /* Now check whether the two intervals are consecutive; if so,
         they can thus be merged */
      if ((sourceInt.begin > IPAddress() && sourceInt.begin - 1 == destInt.end) ||
          (destInt.begin > IPAddress() && destInt.begin - 1 == sourceInt.end)) {
         // One of the two intervals follows the other
         newInterval.begin = (destInt.begin < sourceInt.begin ? destInt.begin : sourceInt.begin);
         newInterval.end = (sourceInt.end > destInt.end ? sourceInt.end : destInt.end);
         merged = true;
      }
      /* Check whether the intervals overlap: their union is
         the largest interval */
      if ((sourceInt.begin >= destInt.begin) && (sourceInt.end <= destInt.end)) {
         // destInterval includes sourceInterval
         newInterval = destInt;
         merged = true;
      }
      if ((destInt.begin >= sourceInt.begin) && (destInt.end <= sourceInt.end)) {
         // sourceInterval includes destInterval
         newInterval = sourceInt;
         merged = true;
      }
      // Inserts a new created interval (if any) in the interval set (addrSet)
      if (merged) {
         retSpace.addrSet.addItem(newInterval, newInterval.begin);
         retSpace.addrCount += newInterval.end - newInterval.begin;
         savedSource = true;
         savedDest = true;
      }
   }
   /* If any of the savedSource and savedDest variables is false,
      this means that an IP interval has not been copied yet */
   if (!savedSource) {
      retSpace.addrSet.addItem(sourceInt, sourceInt.begin);
      retSpace.addrCount += sourceInt.end - sourceInt.begin;
   }
   if (!savedDest) {
      retSpace.addrSet.addItem(destInt, destInt.begin);
      retSpace.addrCount += destInt.end - destInt.begin;
   }
   return (retSpace);
}

// Alias for `join'
IPSubSpace& IPSubSpace::operator+(const IPSubSpace& otherSubSet) {
   (*this) = join(otherSubSet);
   return (*this);
}

// += operator overload
IPSubSpace& IPSubSpace::operator+=(const IPSubSpace& otherSubSpace) {
   (*this) = (*this) + otherSubSpace;
   return (*this);
}

// Returns the number of addresses that currently are in the set
unsigned long long int IPSubSpace::addresses() {
   return addrCount;
}

// Returns a list of the addresses in the set
List<IPAddress>& IPSubSpace::listAddresses() {
   IPAddress curAddr;
   IPInterval curInt;
   bool stop;
   List<IPInterval> intervals;
   // Initialize an internal empty list
   addrRetList = List<IPAddress>();
   intervals = addrSet.listItems();
   while (!intervals.isEmpty()) {
      curInt = intervals.pop();
      curAddr = curInt.begin;
      stop = false;
      while (!stop && curAddr <= curInt.end) {
         addrRetList.enqueue(curAddr);
         /* Must be careful when handling circular values (like IP
            addresses are) */
         if (curAddr == IPAddress(255,255,255,255)) {
            stop = true;
         }
         else {
            curAddr += 1;
         }
      }
   }
   return (addrRetList);
}

/* Returns a list of the addresses in the set, in the form address/netmask
   (IP prefix); performs something similar to the CIDR aggregation */
List<IPPrefix>& IPSubSpace::listPrefixes() {
   IPInterval curInt;
   char pow2;
   unsigned long long int capturedAddresses, toBeCaptured;
   List<IPInterval> intervals;
   // Initialize an internal empty list
   prefRetList = List<IPPrefix>();
   intervals = addrSet.listItems();
   while (!intervals.isEmpty()) {
      curInt = intervals.pop();
      capturedAddresses = 0;
      toBeCaptured = curInt.end - curInt.begin;
      while (capturedAddresses < toBeCaptured) {
         /* Find shortest (i.e., with less `1's) netmask that preserves
            interval's starting address and, at the same time, is
            maximal (i.e., groups the largest number of addresses) */
         pow2 = 0;
         while (pow2 <= 32 && curInt.begin.mask(IPNetmask(32 - pow2)) == curInt.begin
                && ((1ULL << pow2) <= toBeCaptured)) {
            pow2++;
         }
         pow2--;
         capturedAddresses += 1ULL << pow2;
         prefRetList.enqueue(IPPrefix(curInt.begin, IPNetmask(32 - pow2)));
         curInt.begin += 1ULL << pow2;
      }
   }
   return (prefRetList);
}

// ************  IP Subspace  ***********
