/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tdiffGraph.cc                                */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "tdiffGraph.h"

// This function just prints out program name
void printHeader(FILE *file) {
   static bool alreadyPrinted = false;
   if (!alreadyPrinted) {
      if (isTerm(file)) {
         fprintf (file, "%stdiffGraph%s 1.3\n", CONSOLE_BOLD, CONSOLE_NORMAL);
      }
      else {
         fprintf (file, "tdiffGraph 1.3\n");
      }
      alreadyPrinted = true;
   }
}

/* This function processes command line parameters and returns
   true if no error is found; false if there is any syntax error */
bool parseArguments(int argc, char **argv, cmdLineParameters &params, char *error) {
   bool parsed, expectOutFile, expectSourceFile, changed, expectComment, commentSet,
        gotSource;
   short int parsedFileNames, parsedStdins;
   char *fileName;
   int curCond;
   // Default options
   params.listNodes = false;
   params.negate = false;
   params.diffGraph = false;
   params.multiGraph = false;
   params.help = false;
   params.force = false;
   expectOutFile = expectSourceFile = expectComment = commentSet = false;
   gotSource = true;
   parsedFileNames = parsedStdins = 0;
   for (int i = 1; i < argc; i++) {
      parsed = false;
      if (expectOutFile) {
         // This is an output file name
         params.outFile = new char[strlen(argv[i]) + 1];
         strcpy (params.outFile, argv[i]);
         expectOutFile = false;
         parsed = true;
      }
      if (expectSourceFile) {
         // This is a source file name (after a -s option)
         fileName = new char[strlen(argv[i]) + 1];
         strcpy(fileName, argv[i]);
         params.graphFileName.enqueue(fileName);
         parsedFileNames++;
         if (strcmp(fileName, "-") == 0) {
            parsedStdins++;
         }
         gotSource = true;
         expectSourceFile = false;
         parsed = true;
      }
      if (expectComment) {
         // This is a description string for the graph
         strncpy(params.comment, argv[i], MAX_COMMENT_CHARS);
         commentSet = true;
         expectComment = false;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-o") == 0) {
         expectOutFile = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-n") == 0) {
         if (params.negate) {
            strcat (error, "Graph conditions cannot be negated twice");
            return false;
         }
         params.negate = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--help") == 0) {
         params.help = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--comment") == 0) {
         if (commentSet) {
            strcat (error, "There is more than one comment for the output graph");
            return false;
         }
         expectComment = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-l") == 0) {
         if (params.listNodes) {
            strcat (error, "Option -l is used at least twice");
            return false;
         }
         params.listNodes = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-f") == 0) {
         if (params.force) {
            strcat (error, "Option -f is used at least twice");
            return false;
         }
         params.force = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-s") == 0) {
         expectSourceFile = true;
         gotSource = false;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--multi") == 0) {
         if (params.multiGraph) {
            strcat (error, "Option --multi can only be used once");
            return false;
         }
         params.multiGraph = true;
         parsed = true;
      }
      if (!parsed && strncmp(argv[i], "--comp", 6) == 0) {
         if (params.diffGraph) {
            strcat (error, "--over and --comp options can only be used once in a command line");
            return false;
         }
         if (strlen(argv[i] + 6) < 2 || argv[i][6] != ':') {
            strcat (error, "Conditions expected after --comp option");
            return false;
         }
         params.diffGraph = true;
         curCond = 0;
         changed = false;
         for (short int j = 7; j < (short int) strlen(argv[i]); j++) {
            switch (argv[i][j]) {
               case ',':   // Start a new condition
                           if (!changed) {
                              strcat(error, "Empty condition for complementary graph making");
                              return false;
                           }
                           params.cmpMode.enqueue(curCond);
                           changed = false;
                           curCond = 0;
                           break;
               case 'a':   curCond |= COMP_ANY1ST;
                           changed = true;
                           break;
               case 'u':   curCond |= COMP_UNORIENTED1ST;
                           changed = true;
                           break;
               case 'p':   curCond |= COMP_PEERING1ST;
                           changed = true;
                           break;
               case 'd':   curCond |= COMP_DIRECTED1ST;
                           changed = true;
                           break;
               case 'o':   curCond |= COMP_ORIENTED1ST;
                           changed = true;
                           break;
               case 'A':   curCond |= COMP_ANY2ND;
                           changed = true;
                           break;
               case 'U':   curCond |= COMP_UNORIENTED2ND;
                           changed = true;
                           break;
               case 'P':   curCond |= COMP_PEERING2ND;
                           changed = true;
                           break;
               case 'D':   curCond |= COMP_DIRECTED2ND;
                           changed = true;
                           break;
               case 'O':   curCond |= COMP_ORIENTED2ND;
                           changed = true;
                           break;
               default:    strcat (error, "Invalid condition for complementary graph making: `");
                           sprintf(error, "%s%c'", error, argv[i][j]);
                           return false;
            }
         }
         if (!changed) {
            strcat(error, "Empty condition for complementary graph making");
            return false;
         }
         else {
            params.cmpMode.enqueue(curCond);
         }
         parsed = true;
      }
      if (!parsed && strncmp(argv[i], "--over", 6) == 0) {
         if (params.diffGraph) {
            strcat (error, "--over and --comp options can only be used once in a command line");
            return false;
         }
         if (strlen(argv[i] + 6) < 2 || argv[i][6] != ':') {
            strcat (error, "Conditions expected after --over option");
            return false;
         }
         params.diffGraph = true;
         curCond = 0;
         changed = false;
         for (short int j = 7; j < (short int) strlen(argv[i]); j++) {
            switch (argv[i][j]) {
               case ',':   // Start a new condition
                           if (!changed) {
                              strcat(error, "Empty condition for overlap graph making");
                              return false;
                           }
                           params.mode.enqueue(curCond);
                           changed = false;
                           curCond = 0;
                           break;
               case 'a':   curCond |= OVER_ANY;
                           changed = true;
                           break;
               case 'o':   curCond |= OVER_ORIENT1ST;
                           changed = true;
                           break;
               case 'O':   curCond |= OVER_ORIENT2ND;
                           changed = true;
                           break;
               case 'x':   curCond |= OVER_OPPDIRECTED;
                           changed = true;
                           break;
               case 'y':   curCond |= OVER_DIFFDIRECTED;
                           changed = true;
                           break;
               case 'u':   curCond |= OVER_UNORIENT1ST;
                           changed = true;
                           break;
               case 'U':   curCond |= OVER_UNORIENT2ND;
                           changed = true;
                           break;
               case 'p':   curCond |= OVER_PEER1ST;
                           changed = true;
                           break;
               case 'P':   curCond |= OVER_PEER2ND;
                           changed = true;
                           break;
               case 'd':   curCond |= OVER_NONPEER1ST;
                           changed = true;
                           break;
               case 'D':   curCond |= OVER_NONPEER2ND;
                           changed = true;
                           break;
               case 'w':   curCond |= OVER_WEIGHT;
                           changed = true;
                           break;
               default:    strcat (error, "Invalid condition for overlap graph making: `");
                           sprintf(error, "%s%c'", error, argv[i][j]);
                           return false;
            }
         }
         if (!changed) {
            strcat(error, "Empty condition for overlap graph making");
            return false;
         }
         else {
            params.mode.enqueue(curCond);
         }
         parsed = true;
      }
      if (!parsed) {
         // Whatever has not been parsed yet should be a file name
         parsedFileNames++;
         fileName = new char[strlen(argv[i]) + 1];
         strcpy(fileName, argv[i]);
         params.graphFileName.enqueue(fileName);
         if (strcmp(fileName, "-") == 0) {
            parsedStdins++;
         }
      }
      if (parsedStdins > 1) {
         strcat (error, "Only a single input can be taken from standard input");
         return false;
      }
   }
   if (expectComment && !commentSet) {
      strcat (error, "Comment expected after --comment option");
      return false;
   }
   if (expectSourceFile && !gotSource) {
      strcat (error, "Source file name expected after -s option");
      return false;
   }
   if (expectOutFile && params.outFile == NULL) {
      strcat (error, "File name expected after -o option");
      return false;
   }
   if ((params.listNodes && params.diffGraph) ||
       (params.diffGraph && params.multiGraph) ||
       (params.listNodes && params.multiGraph)) {
      strcat (error, "Options -l, --over, --comp, --multi cannot be used together");
      return false;
   }
   if (params.negate && !params.diffGraph) {
      strcat (error, "Option -n can only be used together with either --over or --comp");
      return false;
   }
   if (!params.help && ((parsedFileNames != 2 && !params.multiGraph) ||
       (parsedFileNames < 2 && params.multiGraph))) {
      strcat (error, "Wrong number of input graphs");
      return false;
   }
   return true;
}

// This function prints an on line help
void printHelp(FILE *file, char *cmdLine) {
   char CON_BOLD[strlen(CONSOLE_BOLD) + 1];
   char CON_UNDERLINE[strlen(CONSOLE_UNDERLINE) + 1];
   char CON_NORMAL[strlen(CONSOLE_NORMAL) + 1];
   if (isTerm(file)) {
      strcpy (CON_BOLD, CONSOLE_BOLD);
      strcpy (CON_UNDERLINE, CONSOLE_UNDERLINE);
      strcpy (CON_NORMAL, CONSOLE_NORMAL);
   }
   else {
      strcpy (CON_BOLD, "");
      strcpy (CON_UNDERLINE, "");
      strcpy (CON_NORMAL, "");
   }
   printHeader(file);
   fprintf (file, "%sTORQUE toolkit%s AS graph compare tool\n", CON_BOLD, CON_NORMAL);
   fprintf (file, "USAGE: %s [options] [mode] [[-o] <outFile>] {[-s] <filename>...}\n\n", cmdLine);
   fprintf (file, "By default, output shows a series of statistics about the differences\n");
   fprintf (file, "between the two provided graphs. Anyway, by using any of the modes listed\n");
   fprintf (file, "below, output can be changed.\n\n");
   fprintf (file, "%sMode%s can be at most one of:\n", CON_UNDERLINE, CON_NORMAL);
   fprintf (file, "  -l       Show a list of the nodes that appear only in the first graph,\n");
   fprintf (file, "           only in the second one and in both of them; the first input\n");
   fprintf (file, "           file name on the command line is assumed to be the `first'\n");
   fprintf (file, "           graph, while the other file name corresponds to the `second'\n");
   fprintf (file, "           graph; output will be a list of node numbers, each one preceded\n");
   fprintf (file, "           by a '<', '>' or '|' symbol if, respectively, the node is found\n");
   fprintf (file, "           to be present only in the first, only in the second or in both\n");
   fprintf (file, "           graphs\n\n");
   fprintf (file, "  --over:<conds>  Build an `overlap' or `differential' graph, which is a\n");
   fprintf (file, "                  graph that has an edge (and, of course, its two terminal\n");
   fprintf (file, "                  nodes) whenever the latter is found to exist in both the\n");
   fprintf (file, "                  source graphs and to satisfy a series of conditions that\n");
   fprintf (file, "                  are listed in the <conds> part of the option; <conds> is\n");
   fprintf (file, "                  a list of comma separated combinations of the characters\n");
   fprintf (file, "                  listed below; each item of this list is logically \"or\"ed\n");
   fprintf (file, "                  with the others, while conditions within a single item\n");
   fprintf (file, "                  must hold simultaneously (logical \"and\"); in other\n");
   fprintf (file, "                  words, <conds> represents a DNF formula. Available\n");
   fprintf (file, "                  conditions are:\n");
   fprintf (file, "                    a:     insert the edge anyway\n");
   fprintf (file, "                    o(O):  insert the edge if it is found to be oriented\n");
   fprintf (file, "                           in the first (second) graph\n");
   fprintf (file, "                    u(U):  insert the edge if it is found to be unoriented\n");
   fprintf (file, "                           in the first (second) graph\n");
   fprintf (file, "                    x:     insert the edge if it is directed in both\n");
   fprintf (file, "                           graphs and the two directions are opposite\n");
   fprintf (file, "                    y:     insert the edge if it is oriented in both\n");
   fprintf (file, "                           graphs and the two orientations are different\n");
   fprintf (file, "                    p(P):  insert the edge if it is found to be a peering\n");
   fprintf (file, "                           edge in the first (second) graph\n");
   fprintf (file, "                    d(D):  insert the edge if it is found to be a directed\n");
   fprintf (file, "                           edge in the first (second) graph\n");
   fprintf (file, "                    w:     insert the edge if its weight in the two graphs\n");
   fprintf (file, "                           differs\n");
   fprintf (file, "                  Overlap graph is made up of unoriented edges only.\n\n");
   fprintf (file, "  --comp:<conds>  This option is very similar to the previous one, but\n");
   fprintf (file, "                  this time edges are inserted whenever they only exist\n");
   fprintf (file, "                  in one of the graphs (that is why the generated graph\n");
   fprintf (file, "                  is called `complementary') and they have the properties\n");
   fprintf (file, "                  listed in <conds>; the format of the <conds> list is\n");
   fprintf (file, "                  identical to that of the previous option, but the\n");
   fprintf (file, "                  conditions themselves are a little different:\n");
   fprintf (file, "                    a(A):  insert the edge if it just exists in the first\n");
   fprintf (file, "                           (second) graph\n");
   fprintf (file, "                    u(U):  insert the edge if it is an unoriented edge in\n");
   fprintf (file, "                           the first (second) graph\n");
   fprintf (file, "                    o(O):  insert the edge if it is an oriented edge in\n");
   fprintf (file, "                           the first (second) graph\n");
   fprintf (file, "                    d(D):  insert the edge if it is a directed edge in the\n");
   fprintf (file, "                           first (second) graph\n");
   fprintf (file, "                    p(P):  insert the edge if it is a peering edge in the\n");
   fprintf (file, "                           first (second) graph\n");
   fprintf (file, "                  Unlike the overlap graph, the complementary graph is\n");
   fprintf (file, "                  made up of edges which appear exactly as they did in\n");
   fprintf (file, "                  the source graph.\n\n");
   fprintf (file, "  --multi  This option requires at least two input files, but there is no\n");
   fprintf (file, "           upper limit to their number. Files are read one after the\n");
   fprintf (file, "           other, in the order in which they are provided on command\n");
   fprintf (file, "           line. For each of the edges in the input graphs, an unoriented\n");
   fprintf (file, "           edge is inserted in a so called `multi-differential' graph; if\n");
   fprintf (file, "           such edge is found to already exist in the multi-differential\n");
   fprintf (file, "           graph, its attributes are just updated. Attributes concern\n");
   fprintf (file, "           statistics about the \"history\" of that edge (e.g., the\n");
   fprintf (file, "           number of graphs in which it was found to be unoriented). The\n");
   fprintf (file, "           returned graph is the so modified multi-differential graph\n");
   fprintf (file, "           (file format is the same as plain differential graphs).\n\n");
   fprintf (file, "%sOptions%s are:\n", CON_UNDERLINE, CON_NORMAL);
   fprintf (file, "  -n  Negate conditions for --over and --comp options (i.e., insert edges\n");
   fprintf (file, "      if the conditions in the <conds> list do NOT hold)\n");
   fprintf (file, "  -f  Do not ask user confirmation (default behaviour) when overwriting\n");
   fprintf (file, "      existing files\n");
   fprintf (file, "  --comment <string>  Insert into the graph a comment string; by default,\n");
   fprintf (file, "                      graphs are labelled with the name of the files they\n");
   fprintf (file, "                      were generated from\n");
   fprintf (file, "  --help  Show this help\n\n\n");
   fprintf (file, "<filename> is the name of an input file (a graph generated with\n");
   fprintf (file, "tmakeGraph); use -s option to force following argument to be a source file\n");
   fprintf (file, "name whenever ambiguities may arise with other options on the command\n");
   fprintf (file, "line. Input files must be exactly 2, unless --multi option is used (see\n");
   fprintf (file, "above). Any of the input file names can be replaced with a dash ('-'), in\n");
   fprintf (file, "which case input for the corresponding graph will be taken from standard\n");
   fprintf (file, "input instead of a file.\n\n");
   fprintf (file, "By using -o option, any kind of output (but status messages) can be\n");
   fprintf (file, "directed to a file.\n");
}

/* This function compares two edges according to their
   "lexicographic" ordering (of their terminal nodes),
   and returns a value that is less than, equal to or
   more than 0 if, respectively, first edge is found
   to be "smaller", "equal to" or "greater" than the
   second one */
int lexOrder(ASGraph::graph_edge *edge1,
             ASGraph::graph_edge *edge2) {
   if (edge1->from->key < edge2->from->key ||
       (edge1->from->key == edge2->from->key &&
        edge1->to->key < edge2->to->key)) {
      return -1;
   }
   else {
      if (edge1->from->key > edge2->from->key ||
          (edge1->from->key == edge2->from->key &&
           edge1->to->key > edge2->to->key)) {
         return 1;
      }
      else {
         return 0;
      }
   }
}

/* This function compares nodes in the two graphs and returns
   how many of them are present in only one or both of them; if
   asked for, it also prints a list of the nodes that satisfy
   such a property */
void nodeCmp(FILE *file, cmdLineParameters &params, ASGraph &graph1,
             ASGraph &graph2, diff_type &results) {
   List <ASGraph::graph_node*> nodeList1, nodeList2;
   ASGraph::graph_node *curNode1, *curNode2;
   bool goOn1, goOn2, listNodes = params.listNodes;

   results.nodesOnlyInThe1stGraph = 0;
   results.nodesOnlyInThe2ndGraph = 0;
   results.commonNodes = 0;

   // Produce comparisons concerning nodes
   nodeList1 = graph1.listNodes();
   nodeList2 = graph2.listNodes();
   goOn1 = !nodeList1.isEmpty();
   goOn2 = !nodeList2.isEmpty();
   /* Even if the two following values are dummy (because
      either one of the lists is empty), they will not be
      accessed if the corresponding boolean flag (goOnx)
      is false */
   curNode1 = nodeList1.pop();
   curNode2 = nodeList2.pop();
   while (goOn1 && goOn2) {
      if (curNode1->key < curNode2->key) {
         // This node is only in the first graph
         results.nodesOnlyInThe1stGraph++;
         if (listNodes) {
            fprintf (file, "< %i\n", curNode1->key);
         }
         delete curNode1;
         goOn1 = !nodeList1.isEmpty();
         curNode1 = nodeList1.pop();
      }
      else {
         if (curNode1->key > curNode2->key) {
            // This node is only in the second graph
            results.nodesOnlyInThe2ndGraph++;
            if (listNodes) {
               fprintf (file, "> %i\n", curNode2->key);
            }
            delete curNode2;
            goOn2 = !nodeList2.isEmpty();
            curNode2 = nodeList2.pop();
         }
         else {
            // This node is present in both graphs
            results.commonNodes++;
            if (listNodes) {
               fprintf (file, "| %i\n", curNode1->key);
            }
            delete curNode1;
            goOn1 = !nodeList1.isEmpty();
            curNode1 = nodeList1.pop();
            delete curNode2;
            goOn2 = !nodeList2.isEmpty();
            curNode2 = nodeList2.pop();
         }
      }
   }
   while (goOn1) {
      /* These are, of course, nodes that are only in
         the first graph */
      results.nodesOnlyInThe1stGraph++;
      if (listNodes) {
         fprintf (file, "< %i\n", curNode1->key);
      }
      delete curNode1;
      goOn1 = !nodeList1.isEmpty();
      curNode1 = nodeList1.pop();
   }
   while (goOn2) {
      /* These are, of course, nodes that are only in
         the second graph */
      results.nodesOnlyInThe2ndGraph++;
      if (listNodes) {
         fprintf (file, "> %i\n", curNode2->key);
      }
      delete curNode2;
      goOn2 = !nodeList2.isEmpty();
      curNode2 = nodeList2.pop();
   }
}

/* This function updates mean orientation durations on
   a multi-differential graph; it should be used after
   processing the last graph */
void updateMeans (ASGraph &graph) {
   List<ASGraph::graph_edge*> edgeList;
   ASGraph::graph_edge *curEdge;

   edgeList = graph.listEdges();
   while (!edgeList.isEmpty()) {
      curEdge = edgeList.pop();
      if (curEdge->weight_from->changes == 0) {
         // Current edge never changed its orientation
         curEdge->weight_from->stability = curEdge->weight_to->stability =
            curEdge->weight_from->snapshots;
      }
      else {
         // Current edge changed its orientation at least once
         curEdge->weight_from->stability = curEdge->weight_to->stability =
            (curEdge->weight_from->stability * curEdge->weight_from->changes +
            curEdge->weight_from->snapshots) / (curEdge->weight_from->changes + 1);
         /* Number of changes should not be updated
            (end of processing does not mean edge
            orientation has changed...) */
      }
      delete curEdge->from;
      delete curEdge->to;
      delete curEdge;
   }
}

/* This function alters an existing 'base' (differential) graph by updating
   its "statistic" fields with information coming from other graphs; the
   function should be invoked for each graph that must be compared, but
   always using the same 'base' graph */
void graphCompare(ASGraph &multiGraph, ASGraph &graph) {
   List<ASGraph::graph_edge*> edgeList;
   ASGraph::graph_edge *curEdge, *multiGraphEdge, *multiGraphEdge2;
   ASEdgeWeight newWeight1, newWeight2;

   edgeList = graph.listEdges();
   while (!edgeList.isEmpty()) {
      curEdge = edgeList.pop();
      multiGraphEdge = multiGraph.edgeProperties(curEdge->from->key, curEdge->to->key);
      if (multiGraphEdge != NULL) {
         // Base graph also owns this edge, so it must be updated
         multiGraphEdge->weight_from->graphs++;
         multiGraphEdge->weight_to->graphs++;
         if (curEdge->weight_from->oriented) {
            switch (*(curEdge->orientation)) {
               case DIR_IN:   multiGraphEdge->weight_from->incoming++;
                              multiGraphEdge->weight_to->outgoing++;
                              break;
               case DIR_OUT:  multiGraphEdge->weight_from->outgoing++;
                              multiGraphEdge->weight_to->incoming++;
                              break;
               case DIR_NONE: multiGraphEdge->weight_from->none++;
                              multiGraphEdge->weight_to->none++;
                              break;
            }
         }
         else {
            multiGraphEdge->weight_from->undir++;
            multiGraphEdge->weight_to->undir++;
         }

         if (*(curEdge->orientation) !=
             multiGraphEdge->weight_from->orientation[CMPGRAPH_FIRST] ||
             curEdge->weight_from->oriented !=
             multiGraphEdge->weight_from->diffOriented[CMPGRAPH_FIRST]) {

            // Edge orientation has changed: store new orientation...
            multiGraphEdge->weight_from->orientation[CMPGRAPH_FIRST] =
               *(curEdge->orientation);
            multiGraphEdge->weight_from->diffOriented[CMPGRAPH_FIRST] =
            multiGraphEdge->weight_to->diffOriented[CMPGRAPH_FIRST] =
               curEdge->weight_from->oriented;
            multiGraphEdge->weight_to->orientation[CMPGRAPH_FIRST] =
               reverse(*(curEdge->orientation));

            // ...and update statistics
            if (multiGraphEdge->weight_from->changes == 0) {
               // This is the first change
               multiGraphEdge->weight_from->stability = multiGraphEdge->weight_to->stability =
                  multiGraphEdge->weight_from->snapshots;
            }
            else {
               // This is not the first change
               multiGraphEdge->weight_from->stability = multiGraphEdge->weight_to->stability =
                  (multiGraphEdge->weight_from->stability * multiGraphEdge->weight_from->changes +
                  multiGraphEdge->weight_from->snapshots) / (multiGraphEdge->weight_from->changes + 1);
            }
            multiGraphEdge->weight_from->snapshots = multiGraphEdge->weight_to->snapshots = 1;
            multiGraphEdge->weight_from->changes++;
            multiGraphEdge->weight_to->changes++;
         }
         else {
            multiGraphEdge->weight_from->snapshots++;
            multiGraphEdge->weight_to->snapshots++;
         }

         // Update the same edge seen from the other side
         multiGraphEdge2 = multiGraph.edgeProperties(curEdge->to->key, curEdge->from->key);

         // Just copy attributes seen on the other side of the edge
         multiGraphEdge2->weight_from->graphs = multiGraphEdge2->weight_to->graphs =
            multiGraphEdge->weight_from->graphs;
         multiGraphEdge2->weight_from->incoming = multiGraphEdge->weight_from->outgoing;
         multiGraphEdge2->weight_from->outgoing = multiGraphEdge->weight_from->incoming;
         multiGraphEdge2->weight_from->none = multiGraphEdge->weight_from->none;
         multiGraphEdge2->weight_from->undir = multiGraphEdge->weight_from->undir;
         multiGraphEdge2->weight_to->incoming = multiGraphEdge->weight_to->outgoing;
         multiGraphEdge2->weight_to->outgoing = multiGraphEdge->weight_to->incoming;
         multiGraphEdge2->weight_to->none = multiGraphEdge->weight_to->none;
         multiGraphEdge2->weight_to->undir = multiGraphEdge->weight_to->undir;

         multiGraphEdge2->weight_from->diffOriented[CMPGRAPH_FIRST] =
         multiGraphEdge2->weight_to->diffOriented[CMPGRAPH_FIRST] =
            multiGraphEdge->weight_from->diffOriented[CMPGRAPH_FIRST];

         multiGraphEdge2->weight_from->orientation[CMPGRAPH_FIRST] =
            reverse(multiGraphEdge->weight_from->orientation[CMPGRAPH_FIRST]);
         multiGraphEdge2->weight_to->orientation[CMPGRAPH_FIRST] =
            multiGraphEdge->weight_from->orientation[CMPGRAPH_FIRST];
         multiGraphEdge2->weight_from->snapshots = multiGraphEdge2->weight_to->snapshots =
            multiGraphEdge->weight_from->snapshots;
         multiGraphEdge2->weight_from->stability = multiGraphEdge2->weight_to->stability =
            multiGraphEdge->weight_from->stability;
         multiGraphEdge2->weight_from->changes = multiGraphEdge2->weight_to->changes =
            multiGraphEdge->weight_from->changes;

         delete multiGraphEdge->from;
         delete multiGraphEdge->to;
         delete multiGraphEdge;
         delete multiGraphEdge2->from;
         delete multiGraphEdge2->to;
         delete multiGraphEdge2;
      }
      else {
         // Edge does not exist in multiGraph, so it must be inserted
         multiGraph.addNode(curEdge->from->key, curEdge->from->key);
         multiGraph.addNode(curEdge->to->key, curEdge->to->key);

         // Reset edge weight to default
         newWeight1 = newWeight2 = ASEdgeWeight();
         newWeight1.oriented = newWeight2.oriented = false;
         newWeight1.diffOriented[CMPGRAPH_FIRST] =
         newWeight2.diffOriented[CMPGRAPH_FIRST] =
            curEdge->weight_from->oriented;
         newWeight1.orientation[CMPGRAPH_FIRST] = *(curEdge->orientation);
         newWeight2.orientation[CMPGRAPH_FIRST] = reverse(*(curEdge->orientation));

         if (curEdge->weight_from->oriented) {
            switch (*(curEdge->orientation)) {
               case DIR_NONE: newWeight1.none++;
                              newWeight2.none++;
                              break;
               case DIR_IN:   newWeight1.incoming++;
                              newWeight2.outgoing++;
                              break;
               case DIR_OUT:  newWeight1.outgoing++;
                              newWeight2.incoming++;
                              break;
            }
         }
         else {
            newWeight1.undir++;
            newWeight2.undir++;
         }
         multiGraph.addEdge(curEdge->from->key, curEdge->to->key,
                            DIR_NONE, newWeight1, newWeight2);
      }

      delete curEdge->from;
      delete curEdge->to;
      delete curEdge;
   }
}

/* This function actually inserts an edge inside an
   overlap or complementary graph; it retrieves
   edge properties seen on the two compared graphs
   from the first two arguments */
void addDiffGraphEdge(ASGraph::graph_edge *edge1, ASGraph::graph_edge *edge2,
                      ASGraph &diffGraph) {
   ASEdgeWeight newWeight1, newWeight2;
   char incoming = 0, outgoing = 0, none = 0, undir = 0;
   edge_dir direction;

   // Insert the two terminal nodes (any of them might exist)
   diffGraph.addNode(edge1->from->key, edge1->from->key);
   diffGraph.addNode(edge1->to->key, edge1->to->key);

   // Insert the edge
   newWeight1.pathCovering[CMPGRAPH_FIRST] = edge1->weight_from->pathCovering[CMPGRAPH_FIRST];
   newWeight1.IPCovering[CMPGRAPH_FIRST] = edge1->weight_from->IPCovering[CMPGRAPH_FIRST];
   newWeight2.pathCovering[CMPGRAPH_FIRST] = edge1->weight_to->pathCovering[CMPGRAPH_FIRST];
   newWeight2.IPCovering[CMPGRAPH_FIRST] = edge1->weight_to->IPCovering[CMPGRAPH_FIRST];
   if (edge2 != NULL) {
      // Generating an overlap graph
      newWeight1.oriented = newWeight2.oriented = false;
      newWeight1.diffOriented[CMPGRAPH_FIRST] = newWeight2.diffOriented[CMPGRAPH_FIRST] =
         edge1->weight_from->oriented;
      newWeight1.diffOriented[CMPGRAPH_SECOND] = newWeight2.diffOriented[CMPGRAPH_SECOND] =
         edge2->weight_from->oriented;
      newWeight1.pathCovering[CMPGRAPH_SECOND] = edge2->weight_from->pathCovering[CMPGRAPH_FIRST];
      newWeight1.IPCovering[CMPGRAPH_SECOND] = edge2->weight_from->IPCovering[CMPGRAPH_FIRST];
      newWeight1.orientation[CMPGRAPH_FIRST] = *(edge1->orientation);
      newWeight1.orientation[CMPGRAPH_SECOND] = *(edge2->orientation);
      if (!edge1->weight_from->oriented) {
         undir++;
      }
      else {
         switch (*(edge1->orientation)) {
            case DIR_IN:   incoming++;
                           break;
            case DIR_NONE: none++;
                           break;
            case DIR_OUT:  outgoing++;
                           break;
         }
      }
      if (!edge2->weight_from->oriented) {
         undir++;
      }
      else {
         switch (*(edge2->orientation)) {
            case DIR_IN:   incoming++;
                           break;
            case DIR_NONE: none++;
                           break;
            case DIR_OUT:  outgoing++;
                           break;
         }
      }
      newWeight1.graphs = newWeight2.graphs = 2;
      newWeight1.incoming = newWeight2.outgoing = incoming;
      newWeight1.outgoing = newWeight2.incoming = outgoing;
      newWeight1.none = newWeight2.none = none;
      newWeight1.undir = newWeight2.undir = undir;

      newWeight2.pathCovering[CMPGRAPH_SECOND] = edge2->weight_to->pathCovering[CMPGRAPH_FIRST];
      newWeight2.IPCovering[CMPGRAPH_SECOND] = edge2->weight_to->IPCovering[CMPGRAPH_FIRST];
      newWeight2.orientation[CMPGRAPH_FIRST] = reverse(*(edge1->orientation));
      newWeight2.orientation[CMPGRAPH_SECOND] = reverse(*(edge2->orientation));
      
      // All the edges in an overlap graph appear as unoriented and undirected
      direction = DIR_NONE;
   }
   else {
      /* In the else case, we are generating a complementary graph,
         which appears as a normal graph; so, no other attributes
         need to be be set */
      newWeight1.oriented = newWeight2.oriented = edge1->weight_from->oriented;
      direction = *(edge1->orientation);
   }

   diffGraph.addEdge(edge1->from->key, edge1->to->key, direction,
                     newWeight1, newWeight2);
}

/* This function checks whether an edge conforms to
   a user provided set of conditions; it must be applied
   when building a complementary graph */
bool isCmpGraphEdge(ASGraph::graph_edge *edge, ASGraph &diffGraph,
                    List<int> &conds, bool leftGraph) {
   bool goOn, doAddEdge, tempAddEdge;
   int curCond;
   goOn = !conds.isEmpty();
   doAddEdge = false;
   for (conds.goToFirst(); goOn; goOn = conds.next()) {
      curCond = conds.current();
      tempAddEdge = true;
      if (curCond & COMP_ANY1ST) {
         tempAddEdge = tempAddEdge && leftGraph;
      }
      if (curCond & COMP_UNORIENTED1ST) {
         tempAddEdge = tempAddEdge && leftGraph &&
                    !edge->weight_from->oriented;
      }
      if (curCond & COMP_PEERING1ST) {
         tempAddEdge = tempAddEdge && leftGraph &&
                    edge->weight_from->oriented && *(edge->orientation) == DIR_NONE;
      }
      if (curCond & COMP_DIRECTED1ST) {
         tempAddEdge = tempAddEdge && leftGraph &&
                    edge->weight_from->oriented && *(edge->orientation) != DIR_NONE;
      }
      if (curCond & COMP_ORIENTED1ST) {
         tempAddEdge = tempAddEdge && leftGraph &&
                    edge->weight_from->oriented;
      }
      if (curCond & COMP_ANY2ND) {
         tempAddEdge = tempAddEdge && !leftGraph;
      }
      if (curCond & COMP_UNORIENTED2ND) {
         tempAddEdge = tempAddEdge && !leftGraph &&
                    !edge->weight_from->oriented;
      }
      if (curCond & COMP_PEERING2ND) {
         tempAddEdge = tempAddEdge && !leftGraph &&
                    edge->weight_from->oriented && *(edge->orientation) == DIR_NONE;
      }
      if (curCond & COMP_DIRECTED2ND) {
         tempAddEdge = tempAddEdge && !leftGraph &&
                    edge->weight_from->oriented && *(edge->orientation) != DIR_NONE;
      }
      if (curCond & COMP_ORIENTED2ND) {
         tempAddEdge = tempAddEdge && !leftGraph &&
                    edge->weight_from->oriented;
      }

      doAddEdge = doAddEdge || tempAddEdge;
   }
   return doAddEdge;
}

/* This function checks whether a given couple of
   edges matches user provided conditions (i.e.,
   it tests whether an edge on the overlap graph
   should actually be inserted) */
bool isDiffGraphEdge(ASGraph::graph_edge *edge1, ASGraph::graph_edge *edge2,
                     ASGraph &diffGraph, List<int> &conds) {
   bool goOn, doAddEdge, tempAddEdge;
   int curCond;
   goOn = !conds.isEmpty();
   doAddEdge = false;
   for (conds.goToFirst(); goOn; goOn = conds.next()) {
      curCond = conds.current();
      tempAddEdge = true;
      if (curCond & OVER_ANY) {
         // No changes need to be made to tempAddEdge
      }
      if (curCond & OVER_ORIENT1ST) {
         tempAddEdge = tempAddEdge && edge1->weight_from->oriented;
      }
      if (curCond & OVER_ORIENT2ND) {
         tempAddEdge = tempAddEdge && edge2->weight_from->oriented;
      }
      if (curCond & OVER_UNORIENT1ST) {
         tempAddEdge = tempAddEdge && !edge1->weight_from->oriented;
      }
      if (curCond & OVER_UNORIENT2ND) {
         tempAddEdge = tempAddEdge && !edge2->weight_from->oriented;
      }
      if (curCond & OVER_OPPDIRECTED) {
         tempAddEdge = tempAddEdge &&
                     edge1->weight_from->oriented && edge2->weight_from->oriented &&
                     *(edge1->orientation) == reverse(*(edge2->orientation)) &&
                     *(edge1->orientation) != DIR_NONE;
      }
      if (curCond & OVER_DIFFDIRECTED) {
         tempAddEdge = tempAddEdge &&
                     edge1->weight_from->oriented && edge2->weight_from->oriented &&
                     *(edge1->orientation) != *(edge2->orientation);
      }
      if (curCond & OVER_PEER1ST) {
         tempAddEdge = tempAddEdge && edge1->weight_from->oriented &&
                     *(edge1->orientation) == DIR_NONE;
      }
      if (curCond & OVER_PEER2ND) {
         tempAddEdge = tempAddEdge && edge2->weight_from->oriented &&
                     *(edge2->orientation) == DIR_NONE;
      }
      if (curCond & OVER_NONPEER1ST) {
         tempAddEdge = tempAddEdge && edge1->weight_from->oriented &&
                     *(edge1->orientation) != DIR_NONE;
      }
      if (curCond & OVER_NONPEER2ND) {
         tempAddEdge = tempAddEdge && edge2->weight_from->oriented &&
                     *(edge2->orientation) != DIR_NONE;
      }
      if (curCond & OVER_WEIGHT) {
         tempAddEdge = tempAddEdge &&
                     (edge1->weight_from->pathCovering != edge2->weight_from->pathCovering ||
                     edge1->weight_from->IPCovering != edge2->weight_from->IPCovering);
      }

      doAddEdge = doAddEdge || tempAddEdge;
   }
   return doAddEdge;
}

/* This function accomplishes two operations: it computes numerical differences
   between the edges of the two graphs and it builds a graph corresponding
   to these differences */
void makeDiffGraph(FILE *file, cmdLineParameters &params,
                   ASGraph &graph1, ASGraph &graph2, ASGraph &diffGraph,
                   diff_type &results) {
   List <ASGraph::graph_edge*> edgeList1, edgeList2;
   ASGraph::graph_edge *curEdge1, *curEdge2;
   bool goOn1, goOn2, negate = params.negate, insertEdge;

   results.edgesOnlyInThe1stGraph = 0;
   results.edgesOnlyInThe2ndGraph = 0;
   results.commonEdges = 0;
   results.peersIn1stGraph = results.peersIn2ndGraph = 0;
   results.directedIn1stGraph = results.directedIn2ndGraph = 0;
   results.unorientedIn1stGraph = results.unorientedIn2ndGraph = 0;
   results.diffDirected = 0;
   results.oppDirected = 0;
   results.consDirected = 0;
   results.peersOnlyInThe1stGraph = results.peersOnlyInThe2ndGraph = 0;
   results.directedOnlyInThe1stGraph = results.directedOnlyInThe2ndGraph = 0;
   results.unorientedOnlyInThe1stGraph = results.unorientedOnlyInThe2ndGraph = 0;

   // Produce both comparisons about edges and overlap graph (if requested)
   edgeList1 = graph1.listEdges();
   edgeList2 = graph2.listEdges();
   goOn1 = !edgeList1.isEmpty();
   goOn2 = !edgeList2.isEmpty();
   /* Even if the two following values are dummy (because
      either one of the lists is empty), they will not be
      accessed if the corresponding boolean flag (goOnx)
      is false */
   curEdge1 = edgeList1.pop();
   curEdge2 = edgeList2.pop();
   while (goOn1 && goOn2) {
      if (lexOrder(curEdge1, curEdge2) < 0) {

         // This edge is present only in the first graph
         results.edgesOnlyInThe1stGraph++;
         if (curEdge1->weight_from->oriented) {
            switch (*(curEdge1->orientation)) {
               case DIR_NONE: results.peersIn1stGraph++;
                              results.peersOnlyInThe1stGraph++;
                              break;
               default:       results.directedIn1stGraph++;
                              results.directedOnlyInThe1stGraph++;
                              break;
            }
         }
         else {
            // Current edge is unoriented
            results.unorientedIn1stGraph++;
            results.unorientedOnlyInThe1stGraph++;
         }

         if (!params.cmpMode.isEmpty()) {
            // A complementary graph must be built
            insertEdge = isCmpGraphEdge(curEdge1, diffGraph, params.cmpMode, true);
            if ((insertEdge && !negate) || (!insertEdge && negate)) {
               addDiffGraphEdge(curEdge1, NULL, diffGraph);
            }
         }
         delete curEdge1->from;
         delete curEdge1->to;
         delete curEdge1;
         goOn1 = !edgeList1.isEmpty();
         curEdge1 = edgeList1.pop();
      }
      else {
         if (lexOrder(curEdge1, curEdge2) > 0) {

            // This edge is present only in the second graph
            results.edgesOnlyInThe2ndGraph++;
            if (curEdge2->weight_from->oriented) {
               switch (*(curEdge2->orientation)) {
                  case DIR_NONE: results.peersIn2ndGraph++;
                                 results.peersOnlyInThe2ndGraph++;
                                 break;
                  default:       results.directedIn2ndGraph++;
                                 results.directedOnlyInThe2ndGraph++;
                                 break;
               }
            }
            else {
               // Current edge is unoriented
               results.unorientedIn2ndGraph++;
               results.unorientedOnlyInThe2ndGraph++;
            }

            if (!params.cmpMode.isEmpty()) {
               // A complementary graph must be built
               insertEdge = isCmpGraphEdge(curEdge2, diffGraph, params.cmpMode, false);
               if ((insertEdge && !negate) || (!insertEdge && negate)) {
                  addDiffGraphEdge(curEdge2, NULL, diffGraph);
               }
            }
            delete curEdge2->from;
            delete curEdge2->to;
            delete curEdge2;
            goOn2 = !edgeList2.isEmpty();
            curEdge2 = edgeList2.pop();
         }
         else {

            // This edge is common to both graphs
            results.commonEdges++;
            if (curEdge1->weight_from->oriented) {
               switch (*(curEdge1->orientation)) {
                  case DIR_NONE: results.peersIn1stGraph++;
                                 break;
                  default:       results.directedIn1stGraph++;
                                 break;
               }
            }
            else {
               // Current edge is unoriented in the first graph
               results.unorientedIn1stGraph++;
            }
            if (curEdge2->weight_from->oriented) {
               switch (*(curEdge2->orientation)) {
                  case DIR_NONE: results.peersIn2ndGraph++;
                                 break;
                  default:       results.directedIn2ndGraph++;
                                 break;
               }
            }
            else {
               // Current edge is unoriented in the second graph
               results.unorientedIn2ndGraph++;
            }
            // Trace oppositely and differently directed edges
            if (curEdge1->weight_from->oriented && curEdge2->weight_from->oriented) {
               if (*(curEdge1->orientation) == reverse(*(curEdge2->orientation)) &&
                   *(curEdge1->orientation) != DIR_NONE) {
                  // Edges have opposite orientation
                  results.oppDirected++;
               }
               if (*(curEdge1->orientation) != *(curEdge2->orientation)) {
                  // Edges have different (which can mean opposite) orientation
                  results.diffDirected++;
               }
               if (*(curEdge1->orientation) == *(curEdge2->orientation)) {
                  // Edges are consistently directed
                  results.consDirected++;
               }
            }

            if (!params.mode.isEmpty()) {
               // An overlap graph must be built
               insertEdge = isDiffGraphEdge(curEdge1, curEdge2, diffGraph, params.mode);
               if ((insertEdge && !negate) || (!insertEdge && negate)) {
                  addDiffGraphEdge(curEdge1, curEdge2, diffGraph);
               }
            }
            delete curEdge1->from;
            delete curEdge1->to;
            delete curEdge1;
            goOn1 = !edgeList1.isEmpty();
            curEdge1 = edgeList1.pop();
            delete curEdge2->from;
            delete curEdge2->to;
            delete curEdge2;
            goOn2 = !edgeList2.isEmpty();
            curEdge2 = edgeList2.pop();
         }
      }
   }
   while (goOn1) {
      // This is surely an edge only in the first graph
      results.edgesOnlyInThe1stGraph++;
      if (curEdge1->weight_from->oriented) {
         switch (*(curEdge1->orientation)) {
            case DIR_NONE: results.peersIn1stGraph++;
                           results.peersOnlyInThe1stGraph++;
                           break;
            default:       results.directedIn1stGraph++;
                           results.directedOnlyInThe1stGraph++;
                           break;
         }
      }
      else {
         // Current edge is unoriented
         results.unorientedIn1stGraph++;
         results.unorientedOnlyInThe1stGraph++;
      }
      if (!params.cmpMode.isEmpty()) {
         // A complementary graph must be built
         insertEdge = isCmpGraphEdge(curEdge1, diffGraph, params.cmpMode, true);
         if ((insertEdge && !negate) || (!insertEdge && negate)) {
            addDiffGraphEdge(curEdge1, NULL, diffGraph);
         }
      }
      delete curEdge1->from;
      delete curEdge1->to;
      delete curEdge1;
      goOn1 = !edgeList1.isEmpty();
      curEdge1 = edgeList1.pop();
   }
   while (goOn2) {
      // This is surely an edge only in the second graph
      results.edgesOnlyInThe2ndGraph++;
      if (curEdge2->weight_from->oriented) {
         switch (*(curEdge2->orientation)) {
            case DIR_NONE: results.peersIn2ndGraph++;
                           results.peersOnlyInThe2ndGraph++;
                           break;
            default:       results.directedIn2ndGraph++;
                           results.directedOnlyInThe2ndGraph++;
                           break;
         }
      }
      else {
         // Current edge is unoriented
         results.unorientedIn2ndGraph++;
         results.unorientedOnlyInThe2ndGraph++;
      }
      if (!params.cmpMode.isEmpty()) {
         // A complementary graph must be built
         insertEdge = isCmpGraphEdge(curEdge2, diffGraph, params.cmpMode, false);
         if ((insertEdge && !negate) || (!insertEdge && negate)) {
            addDiffGraphEdge(curEdge2, NULL, diffGraph);
         }
      }
      delete curEdge2->from;
      delete curEdge2->to;
      delete curEdge2;
      goOn2 = !edgeList2.isEmpty();
      curEdge2 = edgeList2.pop();
   }
}

int main(int argc, char **argv) {
   FILE *graphFile1, *graphFile2, *outFile;
   diff_type results;
   ASGraph graph1, graph2, dgraph;
   char error[513];              // One more byte for '\0'
   char *curFileName;
   bool stdinIsTerm, stdoutIsTerm, goOn;
   cmdLineParameters params;

   // Test whether stdin and stdout are connected to a terminal
   stdinIsTerm = isTerm(stdin);
   stdoutIsTerm = isTerm(stdout);

   if (!parseArguments(argc, argv, params, error)) {
      printHeader(stderr);
      fprintf (stderr, "Syntax error: %s\n", error);
      fprintf (stderr, "Type %s --help for more information\n", argv[0]);
      return 1;
   }

   if (params.help) {
      printHelp(stdout, argv[0]);
      return 0;
   }

   if (stdinIsTerm && stdoutIsTerm) {
      printHeader(stdout);
   }

   // Test whether output file exists
   if (params.outFile != NULL && !params.force) {
      outFile = fopen(params.outFile, "r");
      if (outFile != NULL) {
         fclose(outFile);
         if (!askOverwrite(params.outFile)) {
            return 2;
         }
      }
   }

   if (params.outFile != NULL) {
      // Output must be written to file
      outFile = fopen(params.outFile, "w");
      if (outFile == NULL) {
         printHeader(stderr);
         fprintf (stderr, "Unable to open file `%s' for writing\n", params.outFile);
         fprintf (stderr, "(%s)\n", strerror(errno));
         return 3;
      }
   }
   else {
      // Output must be shown on stdout
      outFile = stdout;
   }


   if (params.multiGraph) {
      goOn = true;
      for (params.graphFileName.goToFirst();
           goOn;
           goOn = params.graphFileName.next()) {
         curFileName = params.graphFileName.current();

         // Be sure to clean traces of former graphs
         graph2 = ASGraph();

         if (strcmp(curFileName, "-") == 0) {
            // Get graph from stdin
            if (!graph2.readFromFile(stdin)) {
               printHeader(stderr);
               fprintf (stderr, "Graph from stdin does not appear to be a valid graph:\n");
               fprintf (stderr, "%s\n", graph2.errReason());
               return 3;
            }
         }
         else {
            // Get graph from file
            graphFile2 = fopen(curFileName, "r");
            if (graphFile2 == NULL) {
               printHeader(stderr);
               fprintf (stderr, "Unable to open graph file `%s'\n", curFileName);
               fprintf (stderr, "(%s)\n", strerror(errno));
               return 2;
            }
            if (stdinIsTerm && stdoutIsTerm) {
               printf ("Reading graph file `%s'...\n", curFileName);
            }
            if (!graph2.readFromFile(graphFile2)) {
               printHeader(stderr);
               fprintf (stderr, "File `%s' does not appear to be a valid graph file:\n",
                  curFileName);
               fprintf (stderr, "%s\n", graph2.errReason());
               fclose (graphFile2);
               return 3;
            }
            fclose (graphFile2);
         }

         if (graph2.isDifferential() && stdinIsTerm && stdoutIsTerm) {
            printf ("Warning: Current graph is a differential graph\n");
         }

         // Now update comparisons
         if (stdinIsTerm && stdoutIsTerm) {
            printf ("Updating statistics...\n");
         }
         graphCompare(graph1, graph2);
      }

      // Update mean values on multi-differential
      if (stdinIsTerm && stdoutIsTerm) {
         printf ("Postprocessing multi-differential graph...\n");
      }
      updateMeans(graph1);


      // Now save graph to file
      if (stdinIsTerm && stdoutIsTerm) {
         printf ("Saving graph to file...\n");
      }
      graph1.isDifferential() = true;
      graph1.setComment(params.comment);
      graph1.setSourceFiles(params.graphFileName, false);
      graph1.writeToFile(outFile);
   }
   else {
      // Just make a comparison between two graphs

      // Load the two graph files that are to be compared
      if (strcmp(params.graphFileName.first(), "-") == 0) {
         // First graph must be taken from standard input
         if (!graph1.readFromFile(stdin)) {
            printHeader(stderr);
            fprintf (stderr, "Left input does not appear to be a valid graph:\n");
            fprintf (stderr, "%s\n", graph1.errReason());
            return 3;
         }
      }
      else {
         graphFile1 = fopen(params.graphFileName.first(), "r");
         if (graphFile1 == NULL) {
            printHeader(stderr);
            fprintf (stderr, "Unable to open source file `%s'\n", params.graphFileName.first());
            fprintf (stderr, "(%s)\n", strerror(errno));
            return 2;
         }
         if (stdinIsTerm && stdoutIsTerm) {
            printf ("Reading graph file `%s'...\n", params.graphFileName.first());
         }
         if (!graph1.readFromFile(graphFile1)) {
            printHeader(stderr);
            fprintf (stderr, "File `%s' does not appear to be a valid graph file:\n",
               params.graphFileName.first());
            fprintf (stderr, "%s\n", graph1.errReason());
            fclose (graphFile1);
            return 3;
         }
         fclose (graphFile1);
      }
      if (graph1.isDifferential() && stdinIsTerm && stdoutIsTerm) {
         printf ("Warning: first input is a differential graph\n");
      }

      if (strcmp(params.graphFileName.last(), "-") == 0) {
         // Second graph must be taken from standard input
         if (!graph2.readFromFile(stdin)) {
            printHeader(stderr);
            fprintf (stderr, "Right input does not appear to be a valid graph:\n");
            fprintf (stderr, "%s\n", graph2.errReason());
            return 3;
         }
      }
      else {
         graphFile2 = fopen(params.graphFileName.last(), "r");
         if (graphFile2 == NULL) {
            printHeader(stderr);
            fprintf (stderr, "Unable to open source file `%s'\n", params.graphFileName.last());
            fprintf (stderr, "(%s)\n", strerror(errno));
            return 2;
         }
         if (stdinIsTerm && stdoutIsTerm) {
            printf ("Reading graph file `%s'...\n", params.graphFileName.last());
         }
         if (!graph2.readFromFile(graphFile2)) {
            printHeader(stderr);
            fprintf (stderr, "File `%s' does not appear to be a valid graph file:\n",
               params.graphFileName.last());
            fprintf (stderr, "%s\n", graph2.errReason());
            fclose (graphFile2);
            return 3;
         }
         fclose (graphFile2);
      }
      if (graph2.isDifferential() && stdinIsTerm && stdoutIsTerm) {
         printf ("Warning: second input is a differential graph\n");
      }

      if (stdinIsTerm && stdoutIsTerm) {
         printf ("Computing differences...\n");
      }
      nodeCmp (outFile, params, graph1, graph2, results);
      if (!params.listNodes && !params.diffGraph) {
         fprintf (outFile, "Nodes in 1st graph:  %10u\n", graph1.nodes());
         fprintf (outFile, "- only in 1st graph: %10u\n", results.nodesOnlyInThe1stGraph);
         fprintf (outFile, "Nodes in 2nd graph:  %10u\n", graph2.nodes());
         fprintf (outFile, "- only in 2nd graph: %10u\n", results.nodesOnlyInThe2ndGraph);
         fprintf (outFile, "Common nodes:        %10u\n", results.commonNodes);
      }

      makeDiffGraph (outFile, params, graph1, graph2, dgraph, results);
      if (!params.listNodes && !params.diffGraph) {
         fprintf (outFile, "Edges in 1st graph:  %10u\n", graph1.edges());
         fprintf (outFile, "- peering:           %10u\n", results.peersIn1stGraph);
         fprintf (outFile, "- directed:          %10u\n", results.directedIn1stGraph);
         fprintf (outFile, "- unoriented:        %10u\n", results.unorientedIn1stGraph);
         fprintf (outFile, "- only in 1st graph: %10u\n", results.edgesOnlyInThe1stGraph);
         fprintf (outFile, "  * peering:         %10u\n", results.peersOnlyInThe1stGraph);
         fprintf (outFile, "  * directed:        %10u\n", results.directedOnlyInThe1stGraph);
         fprintf (outFile, "  * unoriented:      %10u\n", results.unorientedOnlyInThe1stGraph);
         fprintf (outFile, "Edges in 2nd graph:  %10u\n", graph2.edges());
         fprintf (outFile, "- peering:           %10u\n", results.peersIn2ndGraph);
         fprintf (outFile, "- directed:          %10u\n", results.directedIn2ndGraph);
         fprintf (outFile, "- unoriented:        %10u\n", results.unorientedIn2ndGraph);
         fprintf (outFile, "- only in 2nd graph: %10u\n", results.edgesOnlyInThe2ndGraph);
         fprintf (outFile, "  * peering:         %10u\n", results.peersOnlyInThe2ndGraph);
         fprintf (outFile, "  * directed:        %10u\n", results.directedOnlyInThe2ndGraph);
         fprintf (outFile, "  * unoriented:      %10u\n", results.unorientedOnlyInThe2ndGraph);
         fprintf (outFile, "Common edges:        %10u\n", results.commonEdges);
         fprintf (outFile, "- oppositely dir.:   %10u\n", results.oppDirected);
         fprintf (outFile, "- differently dir.:  %10u\n", results.diffDirected);
         fprintf (outFile, "- consistently dir.: %10u\n", results.consDirected);
      }
      if (params.diffGraph) {
         // Dump overlap or complementary graph
         if (stdinIsTerm && stdoutIsTerm && params.diffGraph) {
            printf ("Saving graph to file...\n");
         }
         dgraph.isDifferential() = !params.mode.isEmpty();
         dgraph.setComment(params.comment);
         dgraph.setSourceFiles(params.graphFileName, false);
         dgraph.writeToFile(outFile);
      }
   }
}
