/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tviewGraph.cc                                */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "tviewGraph.h"

/* Warning: using this code assumes your compiler supports partial expression
   evaluation (for example, in (A||B), when A is true, B is not evaluated at
   all */

/* This function prints a given AS graph edge; if the graph it
   comes from is a differential graph, and the second argument
   is true, it prints the directions that edge appeared with
   in the two compared graphs */
void printEdge(ASGraph::graph_edge *edge, bool diffMode) {
   printf ("%5i ", edge->from->key);
   if (diffMode) {
      if (edge->weight_from->diffOriented[CMPGRAPH_FIRST]) {
         switch (edge->weight_from->orientation[CMPGRAPH_FIRST]) {
            case DIR_IN:   printf ("<--");
                           break;
            case DIR_OUT:  printf ("-->");
                           break;
            case DIR_NONE: printf ("<->");
                           break;
            default:       break;
         }
      }
      else {
         printf ("---");
      }
      printf (" ");
      if (edge->weight_from->diffOriented[CMPGRAPH_SECOND]) {
         switch (edge->weight_from->orientation[CMPGRAPH_SECOND]) {
            case DIR_IN:   printf ("<--");
                           break;
            case DIR_OUT:  printf ("-->");
                           break;
            case DIR_NONE: printf ("<->");
                           break;
            default:       break;
         }
      }
      else {
         printf ("---");
      }
   }
   else {
      if (edge->weight_from->oriented) {
         switch (*(edge->orientation)) {
            case DIR_IN:   printf ("<--");
                           break;
            case DIR_OUT:  printf ("-->");
                           break;
            case DIR_NONE: printf ("<->");
                           break;
            default:       break;
         }
      }
      else {
         printf ("---");
      }
   }
   printf (" %-5i", edge->to->key);
}

/* This function prints IP covering information for
   a given edge; `whichGraph' tells the function whether
   to get such information from the first or the second of
   the compared graphs; this is senseless if the current
   one is not a differential graph */
void printIPCovering (ASGraph::graph_edge *edge, cmp_graph whichGraph) {
   printf ("   %15llu (", edge->weight_from->IPCovering[whichGraph]);
   if (edge->weight_from->IPCovering[whichGraph] > 0) {
      printf ("~/%3.0f", (float) 32 - (log(edge->weight_from->IPCovering[whichGraph]) / log(2)));
   }
   else {
      printf (" --- ");
   }
   printf (") %15llu (", edge->weight_to->IPCovering[whichGraph]);
   if (edge->weight_to->IPCovering[whichGraph] > 0) {
      printf ("~/%3.0f", (float) 32 - (log(edge->weight_to->IPCovering[whichGraph]) / log(2)));
   }
   else {
      printf (" --- ");
   }
   printf (")");
}

/* This function prints path covering information for
   a given edge; `whichGraph' tells the function whether
   to get such information from the first or the second of
   the compared graphs; this is senseless if the current
   one is not a differential graph */
void printPathCovering (ASGraph::graph_edge *edge, cmp_graph whichGraph) {
   printf ("   %10lu %10lu",
           edge->weight_from->pathCovering[whichGraph],
           edge->weight_to->pathCovering[whichGraph]);
}

/* This function performs all the actions that can be requested
   by means of the `dump' command; see its on-line help for a
   more detailed explanation of what such actions are */
void dumpGraph(command_target target[], long int ccNum,
               ASGraph *graph,
               ConnCompArray sccs, long int scclen,
               ConnCompArray ccs, long int cclen) {
   ASGraph *workOn;
   List<ASGraph::graph_node*> nodeList;
   List<ASGraph::graph_edge*> edgeList;
   ASGraph::graph_node *curNode;
   ASGraph::graph_edge *curEdge;

   if (graph == NULL) {
      fprintf (stderr, "There is no graph to currently work on\n");
      return;
   }

   switch (target[0]) {
      case TARGET_GRAPH:   workOn = graph;
                           break;
      case TARGET_SCC:     if (ccNum < 1 || ccNum > scclen) {
                              fprintf (stderr, "Error: scc out of range [1-%li]\n", scclen);
                              return;
                           }
                           /* The following type cast is harmless: it is only used in order
                              to deal with the whole graph in the same way as with a (strongly)
                              connected component */
                           workOn = (ASGraph*) sccs[ccNum - 1];
                           break;
      case TARGET_CC:      if (ccNum < 1 || ccNum > cclen) {
                              fprintf (stderr, "Error: cc out of range [1-%li]\n", cclen);
                              return;
                           }
                           /* The following type cast is harmless: it is only used in order
                              to deal with the whole graph in the same way as with a (strongly)
                              connected component */
                           workOn = (ASGraph*) ccs[ccNum - 1];
                           break;
      default:             fprintf (stderr, "So, what should I dump?\n");
                           return;
                           break;
   }

   if (target[1] == TARGET_NONE) {
      fprintf (stderr, "So, what should I dump?\n");
   }
   if (target[1] == TARGET_NODES) {
      nodeList = workOn->listNodes();
      while (!nodeList.isEmpty()) {
         curNode = nodeList.pop();
         printf ("%i\n", curNode->key);
         delete curNode;
      }
   }
   else {
      edgeList = workOn->listEdges();
      while (!edgeList.isEmpty()) {
         curEdge = edgeList.pop();
         printEdge(curEdge, target[1] != TARGET_DIFFERENCES && graph->isDifferential());
         switch (target[1]) {
            case TARGET_EDGES:
               /* User asked for a plain list of edges, so
                  nothing more is to be done */
               break;
            case TARGET_IPCOVERING:
               printIPCovering (curEdge);
               if (graph->isDifferential()) {
                  printIPCovering (curEdge, CMPGRAPH_SECOND);
               }
               break;
            case TARGET_PATHCOVERING:
               printPathCovering (curEdge);
               if (graph->isDifferential()) {
                  printPathCovering (curEdge, CMPGRAPH_SECOND);
               }
               break;
            case TARGET_DIFFERENCES:
               printf ("  %5hi %5hu %7.1f    <--: %-5hu    -->: %-5hu    <->: %-5hu    ---: %-5hu",
                       curEdge->weight_from->graphs, curEdge->weight_from->changes,
                       curEdge->weight_from->stability, curEdge->weight_from->incoming,
                       curEdge->weight_from->outgoing, curEdge->weight_from->none,
                       curEdge->weight_from->undir);
                                       break;
            default:                   break;
         }
         printf ("\n");
         delete curEdge->from;
         delete curEdge->to;
         delete curEdge;
      }
   }
}

/* This function prints a distribution that is stored
   inside a BST (each of its nodes keeps count of how
   many items - sccs, ccs, etc. depending on what the
   distribution represents - exist with a number of
   elements - nodes, edges, etc. - given by the node's
   key); the two messages are printed while dumping the
   distribution; if the last argument is true, then each
   value in the distribution is also printed as a base
   2 logarithm */
void printDistribution (BST<unsigned long long int, unsigned long long int> &distr,
                        const char *message1, const char *message2, bool isIPSpace) {
   List<unsigned long long int> itemList, keyList;
   unsigned long long int howManyItems, curValue;

   itemList = distr.listItems();
   keyList = distr.listKeys();
   // The two lists will be processed at the same time
   while (!itemList.isEmpty() && !keyList.isEmpty()) {
      howManyItems = itemList.pop();
      curValue = keyList.pop();
      printf ("%10llu %s %10llu %s", howManyItems, message1, curValue, message2);
      if (isIPSpace) {
         printf (" (");
         if (curValue > 0) {
            printf ("~/%3.0f", (float) 32 - (log(curValue) / log(2)));
         }
         else {
            printf (" --- ");
         }
         printf (")");
      }
      printf ("\n");
   }
}

/* This function is identical to the previous one, except
   for the fact that it prints distributions of floating
   point values as percent values */
void printDistribution (BST<unsigned long long int, double> &distr,
                        const char *message1, const char *message2) {
   List<unsigned long long int> itemList;
   List<double> keyList;
   unsigned long long int howManyItems;
   double curValue;

   itemList = distr.listItems();
   keyList = distr.listKeys();
   // The two lists will be processed at the same time
   while (!itemList.isEmpty() && !keyList.isEmpty()) {
      howManyItems = itemList.pop();
      curValue = keyList.pop();
      printf ("%10llu %s %12.8f %% %s\n", howManyItems, message1, curValue * 100, message2);
   }
}

/* This function is used when executing a `show' command;
   see its on-line help for a more detailed explanation
   of what it can actually do */
void showCount(command_target target[], long int ccNum,
               char *currentFile, ASGraph *graph,
               ConnCompArray sccs, long int scclen,
               ConnCompArray ccs, long int cclen) {
   ASGraph *workOn;
   BST<unsigned long long int, unsigned long long int> distribution;
   BST<unsigned long long int, double> floatDistribution;
   List<unsigned long long int*> distributionItems;
   List<ASGraph::graph_node*> nodeList;
   List<ASGraph::graph_edge*> edgeList;
   List<char*> sourceFiles;
   ASGraph::graph_edge *curEdge, *smallestEdge, *largestEdge;
   long int smallestIndex, largestIndex;
   unsigned long int curValue, smallestValue, largestValue;
   unsigned long long int *oldValue, samples;
   double meanCovering, lowestValue, highestValue, curFloatValue;

   if (graph == NULL) {
      fprintf (stderr, "There is no graph to currently work on\n");
      return;
   }

   if (target[0] == TARGET_SCCS) {
      if (target[1] == TARGET_INDEX) {
         for (long int i = 0; i < scclen; i++) {
            printf ("SCC %10li has %10u nodes and %10u edges\n", i + 1, sccs[i]->nodes(), sccs[i]->edges());
         }
         return;
      }
      printf ("Strongly connected components are %li\n", scclen);
      smallestIndex = largestIndex = 0;
      smallestValue = (target[1] == TARGET_NODES ? sccs[0]->nodes() : sccs[0]->edges());
      largestValue = smallestValue;
      for (long int i = 0; i < scclen; i++) {
         curValue = (target[1] == TARGET_NODES ? sccs[i]->nodes() : sccs[i]->edges());
         if (curValue > largestValue) {
            // This is (one of) the largest strongly connected component(s)
            largestIndex = i;
            largestValue = curValue;
         }
         if (curValue < smallestValue) {
            // This is (one of) the smallest strongly connected component(s)
            smallestIndex = i;
            smallestValue = curValue;
         }
         oldValue = distribution.findItem((unsigned long long int) curValue);
         if (oldValue != NULL) {
            // This value already exists in the distribution
            (*oldValue)++;
         }
         else {
            // This is a new value in the distribution
            distribution.addItem(1LLU, (unsigned long long int) curValue);
         }
      }
      // Now print the distribution
      printDistribution (distribution, "SCCs with",
                         (target[1] == TARGET_NODES ? "nodes" : "edges"), false);
      if (scclen > 0) {
         printf ("Index of the first largest SCC: %li\n", largestIndex + 1);
         printf ("Index of the first smallest SCC: %li\n", smallestIndex + 1);
      }
      return;
   }
   else {
      if (target[0] == TARGET_CCS) {
         if (target[1] == TARGET_INDEX) {
            for (long int i = 0; i < cclen; i++) {
               printf ("CC %10li has %10u nodes and %10u edges\n", i + 1, ccs[i]->nodes(), ccs[i]->edges());
            }
            return;
         }
         printf ("Connected components are %li\n", cclen);
         smallestIndex = largestIndex = 0;
         smallestValue = (target[1] == TARGET_NODES ? ccs[0]->nodes() : ccs[0]->edges());
         largestValue = smallestValue;
         for (long int i = 0; i < cclen; i++) {
            curValue = (target[1] == TARGET_NODES ? ccs[i]->nodes() : ccs[i]->edges());
            if (curValue > largestValue) {
               // This is (one of) the largest connected component(s)
               largestIndex = i;
               largestValue = curValue;
            }
            if (curValue < smallestValue) {
               // This is (one of) the smallest connected component(s)
               smallestIndex = i;
               smallestValue = curValue;
            }
            oldValue = distribution.findItem((unsigned long long int) curValue);
            if (oldValue != NULL) {
               // This value already exists in the distribution
               (*oldValue)++;
            }
            else {
               // This is a new value in the distribution
               distribution.addItem(1LLU, (unsigned long long int) curValue);
            }
         }
         // Now print the distribution
         printDistribution (distribution, "CCs with",
                            (target[1] == TARGET_NODES ? "nodes" : "edges"), false);
         if (cclen > 0) {
            printf ("Index of the first largest CC: %li\n", largestIndex + 1);
            printf ("Index of the first smallest CC: %li\n", smallestIndex + 1);
         }
         return;
      }
      else {
         switch (target[0]) {
            case TARGET_GRAPH:   workOn = graph;
                                 break;
            case TARGET_SCC:     if (ccNum < 1 || ccNum > scclen) {
                                    fprintf (stderr, "Error: scc out of range [1-%li]\n", scclen);
                                    return;
                                 }
                                 /* The following type cast is harmless: it is only used in order
                                    to deal with the whole graph in the same way as with a (strongly)
                                    connected component */
                                 workOn = (ASGraph*) sccs[ccNum - 1];
                                 break;
            case TARGET_CC:      if (ccNum < 1 || ccNum > cclen) {
                                    fprintf (stderr, "Error: cc out of range [1-%li]\n", cclen);
                                    return;
                                 }
                                 /* The following type cast is harmless: it is only used in order
                                    to deal with the whole graph in the same way as with a (strongly)
                                    connected component */
                                 workOn = (ASGraph*) ccs[ccNum - 1];
                                 break;
            default:             fprintf (stderr, "So, what should I show?\n");
                                 return;
                                 break;
         }
      }
   }

   switch (target[1]) {

      case TARGET_NONE:
         fprintf (stderr, "So, what should I show?\n");
         break;

      case TARGET_NODES:
         printf ("%u\n", workOn->nodes());
         break;

      case TARGET_EDGES:
         printf ("%u\n", workOn->edges());
         break;

      case TARGET_INFO:
         printf ("Graph file: %s\n", currentFile);
         printf ("Graph type: %s\n", (graph->isDifferential() ?
                 "(multi)differential" : "normal"));
         printf ("Graph comment: %s\n", graph->getComment());
         sourceFiles = graph->getSourceFiles();
         printf ("Graph has been generated using %llu files:\n", sourceFiles.len());
         while (!sourceFiles.isEmpty()) {
            printf (" %s\n", sourceFiles.pop());
         }
         break;

      case TARGET_IPCOVERING:
      case TARGET_PATHCOVERING:
         edgeList = workOn->listEdges();
         meanCovering = samples = 0;
         smallestEdge = largestEdge = (edgeList.isEmpty() ? NULL : edgeList.first());
         if (target[1] == TARGET_IPCOVERING) {
            smallestValue = (edgeList.isEmpty() ? 0 :
                             edgeList.first()->weight_from->IPCovering[CMPGRAPH_FIRST]);
         }
         else {
            smallestValue = (edgeList.isEmpty() ? 0 :
                             edgeList.first()->weight_from->pathCovering[CMPGRAPH_FIRST]);
         }
         largestValue = smallestValue;
         while (!edgeList.isEmpty()) {
            curEdge = edgeList.pop();
            if (target[1] == TARGET_IPCOVERING) {
               /* This distribution does not take into account the direction
                  in which addresses were announced over an AS adjacency: just
                  the edge 'importance' is taken care of */
               curValue = curEdge->weight_from->IPCovering[CMPGRAPH_FIRST] +
                        curEdge->weight_to->IPCovering[CMPGRAPH_FIRST];
            }
            else {
               // Same comment as above
               curValue = curEdge->weight_from->pathCovering[CMPGRAPH_FIRST] +
                        curEdge->weight_to->pathCovering[CMPGRAPH_FIRST];
            }
            // Update mean covering value
            meanCovering = (meanCovering * (double) samples + (double) curValue) / (double) ++samples;
            if (curValue > largestValue) {
               // This is (one of) the most covered edges
               if (largestEdge != NULL && largestEdge != smallestEdge) {
                  delete largestEdge->from;
                  delete largestEdge->to;
                  delete largestEdge;
               }
               largestEdge = curEdge;
               largestValue = curValue;
            }
            if (curValue < smallestValue) {
               // This is (one of) the less covered edges
               if (smallestEdge != NULL && smallestEdge != largestEdge) {
                  delete smallestEdge->from;
                  delete smallestEdge->to;
                  delete smallestEdge;
               }
               smallestEdge = curEdge;
               smallestValue = curValue;
            }
            oldValue = distribution.findItem((unsigned long long int) curValue);
            if (oldValue != NULL) {
               // This value already exists in the distribution
               (*oldValue)++;
            }
            else {
               // This is a new value in the distribution
               distribution.addItem(1LLU, (unsigned long long int) curValue);
            }
            if (curEdge != smallestEdge && curEdge != largestEdge) {
               delete curEdge->from;
               delete curEdge->to;
               delete curEdge;
            }
         }
         // Now print the distribution
         printDistribution (distribution,
                            (target[1] == TARGET_IPCOVERING ? "adjacencies with IP covering" :
                             "adjacencies with path covering"), "", target[1] == TARGET_IPCOVERING);
         if (largestEdge != NULL && smallestEdge != NULL) {
            printf ("First most covered adjacency: ");
            printEdge(largestEdge, false);
            printf ("\n");
            printf ("First least covered adjacency: ");
            printEdge(smallestEdge, false);
            printf ("\n");
            delete smallestEdge->from;
            delete smallestEdge->to;
            delete smallestEdge;
            if (smallestEdge != largestEdge) {
               delete largestEdge->from;
               delete largestEdge->to;
               delete largestEdge;
            }
         }
         if (samples > 0) {
            if (target[1] == TARGET_IPCOVERING) {
               printf ("Average IP covering: %17.2f  (", meanCovering);
               if (meanCovering > 0) {
                  printf ("~/%3.0f", (float) 32 - (log(meanCovering) / log(2)));
               }
               else {
                  printf (" --- ");
               }
               printf (")\n");
            }
            else {
               printf ("Average path covering: %17.2f\n", meanCovering);
            }
         }
         break;

      case TARGET_DIFFERENCES:
         edgeList = workOn->listEdges();
         smallestEdge = largestEdge = (edgeList.isEmpty() ? NULL : edgeList.first());
         if (edgeList.isEmpty()) {
            smallestValue = largestValue = 0;
            lowestValue = highestValue = 0;
         }
         else {
            /* Initialize temporary maximum and minimum values to
               those of the first encountered edge */
            curEdge = edgeList.first();
            switch (target[2]) {
               case TARGET_ATTENDANCES:
                  smallestValue = largestValue = curEdge->weight_from->graphs;
                  break;
               case TARGET_DIRECTED:
                  lowestValue = highestValue = (double) (curEdge->weight_from->incoming +
                                               curEdge->weight_from->outgoing) /
                                               (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_UNDIRECTED:
                  lowestValue = highestValue = (double) curEdge->weight_from->none /
                                               (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_ORIENTED:
                  lowestValue = highestValue = (double) (curEdge->weight_from->incoming +
                                               curEdge->weight_from->outgoing +
                                               curEdge->weight_from->none) /
                                               (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_UNORIENTED:
                  lowestValue = highestValue = (double) curEdge->weight_from->undir /
                                               (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_CHANGES:
                  lowestValue = highestValue = (double) curEdge->weight_from->changes /
                                               (double) (curEdge->weight_from->graphs - 1);
                  break;
               case TARGET_MEANDURATION:
                  lowestValue = highestValue = (double) curEdge->weight_from->stability /
                                               (double) curEdge->weight_from->graphs;
                  break;
               default:
                  break;
            }
         }
         while (!edgeList.isEmpty()) {
            curEdge = edgeList.pop();
            switch (target[2]) {
               case TARGET_ATTENDANCES:
                  curValue = curEdge->weight_from->graphs;
                  break;
               case TARGET_DIRECTED:
                  curFloatValue = (double) (curEdge->weight_from->incoming +
                                  curEdge->weight_from->outgoing) /
                                  (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_UNDIRECTED:
                  curFloatValue = (double) curEdge->weight_from->none /
                                  (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_ORIENTED:
                  curFloatValue = (double) (curEdge->weight_from->incoming +
                                  curEdge->weight_from->outgoing +
                                  curEdge->weight_from->none) /
                                  (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_UNORIENTED:
                  curFloatValue = (double) curEdge->weight_from->undir /
                                  (double) curEdge->weight_from->graphs;
                  break;
               case TARGET_CHANGES:
                  curFloatValue = (curEdge->weight_from->graphs == 1 ? 0 :
                                   (double) curEdge->weight_from->changes /
                                   (double) (curEdge->weight_from->graphs - 1));
                  break;
               case TARGET_MEANDURATION:
                  curFloatValue = (double) curEdge->weight_from->stability /
                                  (double) curEdge->weight_from->graphs;
                  break;
               default:
                  break;
            }
            if (target[2] == TARGET_ATTENDANCES) {
               /* This value must be inserted in a distribution of
                  integer values */
               if (curValue > largestValue) {
                  largestValue = curValue;
                  if (largestEdge != NULL && largestEdge != smallestEdge) {
                     delete largestEdge->from;
                     delete largestEdge->to;
                     delete largestEdge;
                  }
                  largestEdge = curEdge;
               }
               if (curValue < smallestValue) {
                  smallestValue = curValue;
                  if (smallestEdge != NULL && smallestEdge != largestEdge) {
                     delete smallestEdge->from;
                     delete smallestEdge->to;
                     delete smallestEdge;
                  }
                  smallestEdge = curEdge;
               }
               oldValue = distribution.findItem((unsigned long long int) curValue);
               if (oldValue != NULL) {
                  // This value already exists in the distribution
                  (*oldValue)++;
               }
               else {
                  // This is a new value in the distribution
                  distribution.addItem(1LLU, (unsigned long long int) curValue);
               }
            }
            else {
               /* This value must be inserted in a distribution of
                  floating point values */
               if (curFloatValue > highestValue) {
                  highestValue = curFloatValue;
                  if (largestEdge != NULL && largestEdge != smallestEdge) {
                     delete largestEdge->from;
                     delete largestEdge->to;
                     delete largestEdge;
                  }
                  largestEdge = curEdge;
               }
               if (curFloatValue < lowestValue) {
                  lowestValue = curFloatValue;
                  if (smallestEdge != NULL && smallestEdge != largestEdge) {
                     delete smallestEdge->from;
                     delete smallestEdge->to;
                     delete smallestEdge;
                  }
                  smallestEdge = curEdge;
               }
               oldValue = floatDistribution.findItem(curFloatValue);
               if (oldValue != NULL) {
                  // This value already exists in the distribution
                  (*oldValue)++;
               }
               else {
                  // This is a new value in the distribution
                  floatDistribution.addItem(1LLU, curFloatValue);
               }
            }
            if (curEdge != smallestEdge && curEdge != largestEdge) {
               delete curEdge->from;
               delete curEdge->to;
               delete curEdge;
            }
         }

         // Now print the (correct) distribution
         switch (target[2]) {
            case TARGET_ATTENDANCES:
               printDistribution(distribution, "edges appeared in", "graphs", false);
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First edge with most attendances: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First edge with least attendances: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            case TARGET_DIRECTED:
               printDistribution(floatDistribution, "edges were directed in", "of the graphs they appeared in");
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First most frequently directed edge: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First most rarely directed edge: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            case TARGET_UNDIRECTED:
               printDistribution(floatDistribution, "edges were undirected in", "of the graphs they appeared in");
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First most frequently undirected edge: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First most rarely undirected edge: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            case TARGET_ORIENTED:
               printDistribution(floatDistribution, "edges were oriented in", "of the graphs they appeared in");
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First most frequently oriented edge: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First most rarely oriented edge: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            case TARGET_UNORIENTED:
               printDistribution(floatDistribution, "edges were unoriented in", "of the graphs they appeared in");
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First most frequently unoriented edge: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First most rarely unoriented edge: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            case TARGET_CHANGES:
               printDistribution(floatDistribution, "edges changed their orientation", "of the times they could");
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First most unstable edge: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First most stable edge: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            case TARGET_MEANDURATION:
               printDistribution(floatDistribution, "edges retained their orientation for", "(on avg.) of the graphs they appeared in");
               if (smallestEdge != NULL && largestEdge != NULL) {
                  printf ("First most steady edge: ");
                  printEdge(largestEdge, false);
                  printf ("\n");
                  printf ("First least steady edge: ");
                  printEdge(smallestEdge, false);
                  printf ("\n");
               }
               break;
            default:
               break;
         }
         if (smallestEdge != NULL && largestEdge != NULL) {
            delete largestEdge->from;
            delete largestEdge->to;
            delete largestEdge;
            if (smallestEdge != largestEdge) {
               delete smallestEdge->from;
               delete smallestEdge->to;
               delete smallestEdge;
            }
         }
         break;

      default:
         break;
   }
}

/* This function releases memory that had been previously
   allocated for the graph and its (strongly) connected
   components; it corresponds to the use of the `clear'
   command */
void clearMem(ASGraph **graph,
              ConnCompArray &sccs, long int &scclen,
              ConnCompArray &ccs, long int &cclen) {
   for (long int i = 0; i < scclen; i++) {
      delete sccs[i];
   }
   delete [] sccs;
   sccs = NULL;
   scclen = 0;

   for (long int i = 0; i < cclen; i++) {
      delete ccs[i];
   }
   delete [] ccs;
   ccs = NULL;
   cclen = 0;

   delete *graph;
   *graph = NULL;
}

/* This function saves to a file a (strongly) connected
   component of index `ccNum'; it corresponds to the
   usage of the `save' command */
void saveGraph(char *fileName, char *currentFile,
               ASGraph *graph, command_target target[],
               long int ccNum, ConnCompArray sccs,
               long int scclen,
               ConnCompArray ccs, long int cclen,
               bool forceOverwrite) {
   FILE *outFile;
   List<ConnectedComponent::graph_node*> nodeList;
   List<ConnectedComponent::graph_edge*> edgeList;
   ConnectedComponent::graph_node *curNode;
   ConnectedComponent::graph_edge *curEdge;
   ConnectedComponent *connComp;
   ASGraph toBeSaved;
   List<char*> sourceFiles;
   bool goOn;

   /* Prepare a pointer to the requested (strongly) connected
      component */
   if (target[0] == TARGET_SCC) {
      if (ccNum < 1 || ccNum > scclen) {
         fprintf (stderr, "Error: SCC index out of range [1-%li]\n", scclen);
         return;
      }
      connComp = sccs[ccNum - 1];
   }
   else {
      if (ccNum < 1 || ccNum > cclen) {
         fprintf (stderr, "Error: CC index out of range [1-%li]\n", cclen);
         return;
      }
      connComp = ccs[ccNum - 1];
   }

   /* Create a 'saveable' copy of the connected component
      (the - strongly - connected components extraction
      procedure returns a list of 'plain' graphs, which do
      not have any 'writeToFile' or 'isDifferential'
      method) */
   nodeList = connComp->listNodes();
   while (!nodeList.isEmpty()) {
      curNode = nodeList.pop();
      toBeSaved.addNode(curNode->key, curNode->key);
      delete curNode;
   }
   edgeList = connComp->listEdges();
   while (!edgeList.isEmpty()) {
      curEdge = edgeList.pop();
      toBeSaved.addEdge(curEdge->from->key, curEdge->to->key, *(curEdge->orientation),
                        *(curEdge->weight_from), *(curEdge->weight_to));
      delete curEdge->from;
      delete curEdge->to;
      delete curEdge;
   }
   sourceFiles.enqueue(currentFile);
   toBeSaved.setSourceFiles(sourceFiles, true);

   /* Set the 'saveable' graph's differential attribute to the
      same value as the working graph's: this will allow the
      'writeToFile' method to save the differential information
      too */
   toBeSaved.isDifferential() = graph->isDifferential();

   goOn = true;
   outFile = fopen(fileName, "r");
   if (outFile != NULL) {
      // File already exists
      fclose(outFile);
      goOn = forceOverwrite || askOverwrite(fileName);
   }

   if (goOn) {
      outFile = fopen(fileName, "w");
      if (outFile == NULL) {
         fprintf (stderr, "Unable to open file `%s' for writing\n", fileName);
         fprintf (stderr, "(%s)\n", strerror(errno));
         return;
      }
      if (!toBeSaved.writeToFile(outFile)) {
         fprintf (stderr, "Error: unable to write graph to file:\n");
         fprintf (stderr, toBeSaved.errReason());
      }
      fclose(outFile);
   }
}

/* This function reads a graph file, stores it in memory
   and computes its (strongly) connected components;
   it is used when issuing the `load' command */
void loadGraph(char *fileName, char *currentFile, ASGraph **graph,
               ConnCompArray &sccs, long int &scclen,
               ConnCompArray &ccs, long int &cclen) {
   FILE *graphFile;
   List<Graph<ASnode, ASEdgeWeight, ASnode>*> ccList;
   long int i;

   strcpy(currentFile, "");

   clearMem(graph, sccs, scclen, ccs, cclen);

   graphFile = fopen(fileName, "r");
   if (graphFile == NULL) {
      fprintf (stderr, "Error: unable to open file `%s' for reading\n", fileName);
      fprintf (stderr, "(%s)\n", strerror(errno));
      return;
   }

   *graph = new ASGraph;
   if (!(*graph)->readFromFile(graphFile)) {
      fprintf (stderr, "File `%s' does not appear to be a valid graph file:\n", fileName);
      fprintf (stderr, "%s\n", (*graph)->errReason());
      fclose(graphFile);
      return;
   }
   fclose(graphFile);

   // Now compute strongly connected components
   ccList = (*graph)->findCC(FIND_SCC);

   /* Copy strongly connected components inside the
      strongly connected components array */
   i = 0;
   sccs = new ConnectedComponent*[ccList.len()];
   ccList.goToFirst();
   while (!ccList.isEmpty()) {
      sccs[i] = ccList.pop();
      i++;
   }
   scclen = i;

   // Now compute connected components
   ccList = (*graph)->findCC(FIND_WCC);

   /* Copy connected components inside the connected
      components array */
   i = 0;
   ccs = new ConnectedComponent*[ccList.len()];
   ccList.goToFirst();
   while (!ccList.isEmpty()) {
      ccs[i] = ccList.pop();
      i++;
   }
   cclen = i;

   strcpy(currentFile, fileName);
}

/* This function is invoked when using the `echo' command;
   it either prints out a given message or enables/disables
   command echoing, by changing the value of its second
   argument */
void echoString(char *message, bool &echo) {
   bool newLine;

   if (message == NULL || strlen(message) == 0) {
      return;
   }

   newLine = true;
   if (strcmp(message, "-on") == 0) {
      echo = true;
   }
   else {
      if (strcmp(message, "-off") == 0) {
         echo = false;
      }
      else {
         if (strncmp(message, "-n", 2) == 0) {
            newLine = false;
            message += 3;
         }
         printf ("%s", message);
         if (newLine) {
            printf ("\n");
         }
      }
   }
}

/* This function takes a token (word) as input, and tests
   how many keywords match that token (which can be a command/argument
   in its shortened form); `cmdLevel' is the index of the
   command/argument that is being processed (it is used in order to
   only allow matches for that argument), while `completeToken'
   is filled (on return) with the first matching keyword (in its
   extended version); returned value is the number of
   matching keywords */
int matchingKeywords(char *token, char *completeToken, int cmdLevel) {
   int matches;
   matches = 0;
   strcpy (completeToken, "");
   for (int i = argIndex[cmdLevel - 1]; i < (cmdLevel == 4 ? 28 : argIndex[cmdLevel]); i++) {
      if (strncmp(token, commandString[i], strlen(token)) == 0) {
         matches++;
         if (matches > 1) {
            strcpy (completeToken, "");
         }
         else {
            strcpy (completeToken, commandString[i]);
         }
      }
   }
   return matches;
}

/* This function parses a command line and, if there are
   no errors, runs the requested command; if any error is
   found, a short explanation is printed out; returned value
   is true if issued command was an exit command; false
   otherwise */
bool executeCommand(char *cmdLine, char *currentFile, bool &echo,
                    ASGraph **graph, ConnCompArray &sccs, long int &scclen,
                    ConnCompArray &ccs, long int &cclen,
                    bool forceOverwrite) {
   char oldCmdLine[MAX_COMMAND_LENGTH + 1],
        completeToken[13], *curToken, *fileName, *err;
   bool finished, parsedToken, helpMode, skip, correctSyntax;
   unsigned char parsedTokens;
   command_action whatToDo;
   command_target target[3];
   long int ccNum;
   int cmdLen, parsedChars;

   target[0] = target[1] = target[2] = TARGET_NONE;
   whatToDo = ACTION_NONE;
   ccNum = -1;
   fileName = NULL;
   finished = false;

   /* Keep a copy of the original command line (tokenizing
      it will overwrite some of its parts) */
   strcpy (oldCmdLine, cmdLine);

   if (echo) {
      printf ("%s\n", cmdLine);
   }

   parsedTokens = 0;
   helpMode = false;
   parsedToken = true;
   skip = false;
   cmdLen = strlen(cmdLine);
   curToken = strtok(cmdLine, " ");
   parsedChars = 0;
   // Analyze tokens
   while (parsedToken && !skip && curToken != NULL && (!helpMode || parsedTokens < 2)) {
      parsedTokens++;
      parsedChars += strlen(curToken);
      parsedToken = false;
      if (matchingKeywords(curToken, completeToken, parsedTokens) > 1) {
         fprintf (stderr, "Ambiguous token: `%s'\n", curToken);
         skip = true;
      }
      switch (parsedTokens) {
         case 1:  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_HELP]) == 0) {
                     parsedToken = true;
                     if (helpMode) {
                        fprintf (stderr, "Error: help command used more than once\n");
                     }
                     else {
                        helpMode = true;
                        parsedTokens--;
                     }
                  }
                  if (!skip && (strcmp(completeToken, commandString[argIndex[0] + CMD_EXIT]) == 0 ||
                     strcmp(completeToken, commandString[argIndex[0] + CMD_QUIT]) == 0)) {
                     parsedToken = true;
                     whatToDo = ACTION_QUIT;
                     finished = !helpMode;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_DUMP]) == 0) {
                     parsedToken = true;
                     whatToDo = ACTION_DUMP;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_LOAD]) == 0) {
                     whatToDo = ACTION_LOAD;
                     if (parsedChars < cmdLen) {
                        curToken += strlen(curToken) + 1;
                     }
                     else {
                        curToken = NULL;
                     }
                     if (!helpMode && (curToken == NULL || strlen(curToken) == 0)) {
                        fprintf (stderr, "Error: file name expected\n");
                        skip = true;
                     }
                     else {
                        parsedToken = true;
                        fileName = curToken;
                        skip = true;         // Force exit from cycle
                     }
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_SHOW]) == 0) {
                     parsedToken = true;
                     whatToDo = ACTION_SHOW;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_CLEAR]) == 0) {
                     parsedToken = true;
                     whatToDo = ACTION_CLEAR;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_ECHO]) == 0) {
                     whatToDo = ACTION_ECHO;
                     if (parsedChars < cmdLen) {
                        curToken += strlen(curToken) + 1;
                     }
                     else {
                        curToken = NULL;
                     }
                     parsedToken = true;
                     skip = true;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[0] + CMD_SAVE]) == 0) {
                     parsedToken = true;
                     whatToDo = ACTION_SAVE;
                  }
                  break;
         case 2:  if (!skip && strcmp(completeToken, commandString[argIndex[1] + ARG_GRAPH]) == 0) {
                     parsedToken = true;
                     target[0] = TARGET_GRAPH;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[1] + ARG_SCC]) == 0) {
                     target[0] = TARGET_SCC;
                     curToken = strtok(NULL, " ");
                     if (curToken == NULL || strlen(curToken) == 0) {
                        fprintf (stderr, "Error: SCC index expected\n");
                        skip = true;
                     }
                     else {
                        ccNum = strtol(curToken, &err, 10);
                        if (strlen(err) > 0) {
                           fprintf (stderr, "Error: invalid number for SCC index (`%s')\n", curToken);
                           skip = true;
                        }
                        else {
                           parsedToken = true;
                        }
                     }
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[1] + ARG_SCCS]) == 0) {
                     parsedToken = true;
                     target[0] = TARGET_SCCS;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[1] + ARG_CC]) == 0) {
                     target[0] = TARGET_CC;
                     curToken = strtok(NULL, " ");
                     if (curToken == NULL || strlen(curToken) == 0) {
                        fprintf (stderr, "Error: CC index expected\n");
                        skip = true;
                     }
                     else {
                        ccNum = strtol(curToken, &err, 10);
                        if (strlen(err) > 0) {
                           fprintf (stderr, "Error: invalid number for CC index (`%s')\n", curToken);
                           skip = true;
                        }
                        else {
                           parsedToken = true;
                        }
                     }
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[1] + ARG_CCS]) == 0) {
                     parsedToken = true;
                     target[0] = TARGET_CCS;
                  }
                  if (!skip && whatToDo == ACTION_SAVE &&
                      (target[0] == TARGET_SCC || target[0] == TARGET_CC)) {
                     if (parsedChars < cmdLen) {
                        curToken += strlen(curToken) + 1;
                     }
                     else {
                        curToken = NULL;
                     }
                     if (curToken == NULL || strlen(curToken) == 0) {
                        fprintf (stderr, "Error: file name expected\n");
                        parsedToken = false;
                        skip = true;
                     }
                     else {
                        parsedToken = true;
                        fileName = curToken;
                        skip = true;         // Force exit from cycle
                     }
                  }
                  break;
         case 3:  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_INFO]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_INFO;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_NODES]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_NODES;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_EDGES]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_EDGES;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_IPCOVERING]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_IPCOVERING;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_PATHCOVERING]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_PATHCOVERING;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_DIFFERENCES]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_DIFFERENCES;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[2] + ARG_INDEX]) == 0) {
                     parsedToken = true;
                     target[1] = TARGET_INDEX;
                  }
                  break;
         case 4:  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_ATTENDANCES]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_ATTENDANCES;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_DIRECTED]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_DIRECTED;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_UNDIRECTED]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_UNDIRECTED;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_ORIENTED]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_ORIENTED;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_UNORIENTED]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_UNORIENTED;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_CHANGES]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_CHANGES;
                  }
                  if (!skip && strcmp(completeToken, commandString[argIndex[3] + ARG_MEANDURATION]) == 0) {
                     parsedToken = true;
                     target[2] = TARGET_MEANDURATION;
                  }
                  break;
         default: fprintf (stderr, "Error: too many arguments\n");
                  skip = true;               // Force exit from cycle
                  break;
      }
      if (parsedToken && !skip) {
         curToken = strtok(NULL, " ");
      }
   }

   if (!parsedToken && !skip && !finished) {
      fprintf (stderr, "Error: invalid keyword `%s'\n", curToken);
   }

   if (parsedToken && !finished) {
      correctSyntax = true;
      if (!helpMode) {
         // Check command line correctness
         switch (whatToDo) {
            case ACTION_DUMP:    if (correctSyntax && (target[0] == TARGET_SCCS || target[0] == TARGET_CCS)) {
                                    fprintf (stderr, "Error: dump command cannot work on multiple (S)CCs\n");
                                    correctSyntax = false;
                                 }
                                 if (correctSyntax && (target[1] == TARGET_INFO || target[1] == TARGET_INDEX)) {
                                    fprintf (stderr, "Error: such information can only be requested by using the `show' command\n");
                                    correctSyntax = false;
                                 }
                                 if (correctSyntax && target[2] != TARGET_NONE) {
                                    fprintf (stderr, "Error: too many arguments\n");
                                    correctSyntax = false;
                                 }
                                 break;
            case ACTION_SHOW:    if (correctSyntax && target[1] == TARGET_INFO && target[0] != TARGET_GRAPH) {
                                    fprintf (stderr, "Error: information can only be shown for the whole graph\n");
                                    correctSyntax = false;
                                 }
                                 if (correctSyntax && target[1] == TARGET_INDEX && target[0] != TARGET_SCCS &&
                                     target[0] != TARGET_CCS) {
                                    fprintf (stderr, "Error: (S)CCs indexes can only be listed for all the (S)CCs\n");
                                    correctSyntax = false;
                                 }
                                 if (correctSyntax && (target[0] == TARGET_SCCS || target[0] == TARGET_CCS) &&
                                    (target[1] != TARGET_NODES && target[1] != TARGET_EDGES &&
                                     target[1] != TARGET_INDEX)) {
                                    fprintf (stderr, "Error: can only show distribution or indexes for (S)CCs\n");
                                    correctSyntax = false;
                                 }
                                 if (correctSyntax && target[2] != TARGET_NONE && target[1] != TARGET_DIFFERENCES) {
                                    fprintf (stderr, "Error: such distribution can only be shown for differences\n");
                                    correctSyntax = false;
                                 }
                                 if (correctSyntax && target[1] == TARGET_DIFFERENCES && target[2] == TARGET_NONE) {
                                    fprintf (stderr, "Error: I don't know which differences to show\n");
                                    correctSyntax = false;
                                 }
                                 break;
            case ACTION_CLEAR:   if (correctSyntax && parsedTokens > 1) {
                                    fprintf (stderr, "Error: `clear' command takes no arguments\n");
                                    correctSyntax = false;
                                 }
                                 break;
            case ACTION_SAVE:    if (correctSyntax && target[0] != TARGET_SCC && target[0] != TARGET_CC) {
                                    fprintf (stderr, "Error: `save' command can only be used to save (S)CCs\n");
                                    correctSyntax = false;
                                 }
                                 break;
            default:             break;
         }
      }

      if (helpMode) {
         switch (whatToDo) {
            case ACTION_NONE:
               printf ("\nAvailable commands are:\n");
               for (int i = 0; i < argIndex[1]; i++) {
                  printf ("%s ", commandString[i]);
               }
               printf ("\nCommands can be truncated, if this does not arise ambiguities (e.g.,\n");
               printf ("`d g n' is the same as `dump graph nodes')\n");
               printf ("\nType `help <command>' to get more help about a command.\n\n");
               break;
            case ACTION_DUMP:
               printf ("\nDump command is used to get a list of items inside a graph or its\n");
               printf ("portion. Command syntax is:\n\n");
               printf ("dump <object> <property>\n\n");
               printf ("where\n\n");
               printf (" <object> is the area to be dumped, and can be one of:\n\n");
               printf ("  graph   tells dump to work on the entire previously loaded graph\n");
               printf ("  scc <n> tells dump to work on the strongly connected component of\n");
               printf ("          index <n>\n");
               printf ("  cc <n>  tells dump to work on the connected component of index <n>\n\n");
               printf (" <property> is the type of content to be listed, and can be one of:\n\n");
               printf ("  nodes         is used to get a list of the nodes; output is a plain\n");
               printf ("                list of AS numbers\n");
               printf ("  edges         is used to get a list of the edges; output is in the\n");
               printf ("                form\n");
               printf ("                   A ddd B\n");
               printf ("                where A and B are the two terminals of the edge, and ddd\n");
               printf ("                can be one of the following:\n");
               printf ("                   --- edge has not been assigned an orientation\n");
               printf ("                       (unoriented)\n");
               printf ("                   <-- edge is directed from B to A\n");
               printf ("                   --> edge is directed from A to B\n");
               printf ("                   <-> edge is a (bi-directed) peering edge\n");
               printf ("  ipcovering    is used to get a more detailed list of the edges,\n");
               printf ("                where for each however directed edge A B, the number of\n");
               printf ("                IP addresses that were announced over it is reported;\n");
               printf ("                output is a list of edges with covering information\n");
               printf ("                following each edge: the first value is the number of\n");
               printf ("                addresses that have been announced from A to B, while\n");
               printf ("                the second one is the number of addresses that have\n");
               printf ("                been announced from B to A; for each value, also its\n");
               printf ("                corresponding \"netmask\" coverage is reported\n");
               printf ("  pathcovering  is used to get a list of the edges together with the\n");
               printf ("                number of paths they appeared in; for each however\n");
               printf ("                directed edge A B, two values are reported: the first\n");
               printf ("                one is the number of appearances of the adjacency A B\n");
               printf ("                among the paths; the second one is the number of\n");
               printf ("                appearances of the reversed adjacency B A among the\n");
               printf ("                same paths\n");
               printf ("  differences   is used to obtain information about edges' orientation\n");
               printf ("                history; when using this mode, the plain list of the\n");
               printf ("                edges is enriched with a set of values, which, from\n");
               printf ("                the leftmost to the rightmost, have the following\n");
               printf ("                meaning:\n");
               printf ("                   - number of compared graphs the edge appeared in\n");
               printf ("                   - number of changes in the orientation of the edge\n");
               printf ("                     (including transitions from an oriented state to\n");
               printf ("                     an unoriented one and vice versa)\n");
               printf ("                   - mean edge orientation firmness (i.e., the mean\n");
               printf ("                     number of subsequent graphs in which the edge\n");
               printf ("                     kept its orientation)\n");
               printf ("                   - the following values represent the number of\n");
               printf ("                     times the edge appeared with a given orientation\n");
               printf ("                     (see above for the meaning of the various\n");
               printf ("                     orientation symbols)\n\n");
               printf ("When dealing with differential graphs, most values (including edge\n");
               printf ("orientations) are reported twice, meaning that the first group of\n");
               printf ("values represents the first of the compared graphs, while the second\n");
               printf ("one represents the second of the compared graphs.\n\n");
               break;
            case ACTION_SHOW:
               printf ("\nThe `show' command can be used to get cumulative information about\n");
               printf ("graph properties. Command syntax is:\n\n");
               printf ("show <object> <property> [<diffProperty>]\n\n");
               printf ("where\n\n");
               printf (" <object> is the area to work on, and can be one of:\n\n");
               printf ("  graph          tells show to work on the entire previously loaded\n");
               printf ("                 graph\n");
               printf ("  scc <n>        focuses the attention on the strongly connected\n");
               printf ("                 component of index <n>\n");
               printf ("  cc <n>         tells show to use the connected component of index\n");
               printf ("                 <n> as its target\n");
               printf ("  anyscc         considers every strongly connected component (use it\n");
               printf ("                 if you wish to print a distribution of the strongly\n");
               printf ("                 connected components size)\n");
               printf ("  anycc          works the same as `anyscc', but instead focuses on\n");
               printf ("                 connected components\n\n");
               printf (" <property> is the entity whose size (distribution) should be reported,\n");
               printf (" and can be one of:\n\n");
               printf ("   information   which is the only \"strange\" kind of property; use\n");
               printf ("                 this keyword to print general information concerning\n");
               printf ("                 the graph\n");
               printf ("   nodes         which just reports the number of nodes in the selected\n");
               printf ("                 portion of the graph\n");
               printf ("   edges         which reports the number of edges\n");
               printf ("   index         which lists the indexes of the (S)CCs, together with\n");
               printf ("                 their size; this keyword can only be used when target\n");
               printf ("                 is either `anyscc' or `anycc'\n");
               printf ("   ipcovering    if you want to get a distribution of the number of\n");
               printf ("                 addresses that have been advertised over each edge\n");
               printf ("                 (regardless of the direction in which they were announced\n");
               printf ("                 over it)\n");
               printf ("   pathcovering  to print a distribution of the number of paths each\n");
               printf ("                 edge appeared in (regardless of the order in which its\n");
               printf ("                 end ASs appeared in the paths)\n");
               printf ("   differences   in order to unlock a new series of keywords which can\n");
               printf ("                 be used to get information about graph comparisons\n");
               printf ("                 results\n\n");
               printf (" <diffProperty> is one of the following differential properties (use\n");
               printf (" this argument only when <property> is `differences'):\n\n");
               printf ("   attendances   shows a distribution of the number of compared graphs\n");
               printf ("                 each edge appeared in\n");
               printf ("   directed      shows a distribution of the number of times each edge\n");
               printf ("                 appeared directed with respect to its attendances\n");
               printf ("   undirected    the same as `directed', but for undirected instances\n");
               printf ("   oriented      the same as `directed', but for oriented instances\n");
               printf ("   unoriented    the same as `directed', but for unoriented instances\n");
               printf ("   changes       shows a distribution of the number of orientation\n");
               printf ("                 changes each edge has undergone, with respect to its\n");
               printf ("                 attendances\n");
               printf ("   duration      shows a distribution of the mean duration of an\n");
               printf ("                 orientation - number of subsequent graphs in which the\n");
               printf ("                 edge retained the same orientation - with respect to\n");
               printf ("                 the edge's attendances\n\n");
               break;
            case ACTION_LOAD:
               printf ("\nThe `load' command reads a graph file and stores it in memory after\n");
               printf ("having computed its connected components. From that moment onwards,\n");
               printf ("the graph is available for processing by means of the other commands.\n");
               printf ("Command syntax is:\n\n");
               printf ("load <filename>\n\n");
               printf ("where <filename> is the name of the graph file to be loaded\n\n");
               break;
            case ACTION_CLEAR:
               printf ("\nUse the `clear' command to release memory by deleting the graph you\n");
               printf ("are currently working on. Command syntax is:\n\n");
               printf ("clear\n\n");
               break;
            case ACTION_ECHO:
               printf ("\nThis command can be used both to switch on/off command echoing and to\n");
               printf ("display a user message. Command syntax is:\n\n");
               printf ("echo [-n|-off|-on] <message>\n\n");
               printf ("where:\n");
               printf (" -n   tells echo not to insert a newline character after displaying\n");
               printf ("      <message>\n");
               printf (" -off tells echo to disable command echoing (no message must be given)\n");
               printf (" -on  tells echo to enable command echoing (no message must be given)\n\n");
               break;
            case ACTION_SAVE:
               printf ("\nThe `save' command allows to isolate a (strongly) connected component\n");
               printf ("of the currently processed graph and to save it to a file (which can\n");
               printf ("then be loaded for processing). Command syntax is:\n\n");
               printf ("save {scc <n>|cc <n>} <filename>\n\n");
               printf ("where:\n");
               printf ("  scc and cc are used to specify which (strongly) connected component\n");
               printf ("             to work on\n");
               printf ("  <filename> is the name of the file to save the extracted component\n");
               printf ("             to\n\n");
               break;
            case ACTION_QUIT:
               printf ("\nBoth `exit' and `quit' are used to clean up memory and abandon the\n");
               printf ("program. Command syntaxes are:\n\n");
               printf ("exit\n");
               printf ("quit\n\n");
               break;
         }
         return false;
      }

      if (correctSyntax) {
         switch(whatToDo) {
            case ACTION_DUMP:    dumpGraph(target, ccNum, *graph, sccs, scclen, ccs, cclen);
                                 break;
            case ACTION_SHOW:    showCount(target, ccNum, currentFile, *graph, sccs, scclen, ccs, cclen);
                                 break;
            case ACTION_LOAD:    loadGraph(fileName, currentFile, graph, sccs, scclen, ccs, cclen);
                                 break;
            case ACTION_CLEAR:   clearMem(graph, sccs, scclen, ccs, cclen);
                                 break;
            case ACTION_ECHO:    echoString(curToken, echo);
                                 break;
            case ACTION_SAVE:    saveGraph(fileName, currentFile, *graph, target, ccNum, sccs, scclen,
                                          ccs, cclen, forceOverwrite);
                                 break;
            default:             break;
         }
      }
   }

   return finished;
}

/* The following function just starts a cycle that
   allows user to issue commands by using an on-line
   prompt */
void goInteractive(const cmdLineParameters &params) {
   char cmdLine[MAX_COMMAND_LENGTH + 1],
        curCommand[MAX_COMMAND_LENGTH + 1], *curCmd;
   bool echo = false, quit;
   ASGraph *graph = NULL;
   ConnCompArray ccs, sccs;
   long int scclen, cclen;
   char currentFile[513];     // One more byte for `\0'
   int parsedLen, cmdLen;

   ccs = sccs = NULL;
   scclen = cclen = 0;

   printf ("Interactive mode; type `help' to get a list of available commands\n");

   do {
      parsedLen = 0;
      quit = false;

      printf ("tvg> ");
      fflush (stdout);
      fgets(cmdLine, MAX_COMMAND_LENGTH, stdin);

      // Remove CR/LF at end of line
      if (strpbrk(cmdLine, "\n\r")) {
         *(strpbrk(cmdLine, "\n\r")) = '\0';
      }

      // Remove comments
      if (cmdLine[0] == '#') {
         // The whole line is a comment
         strcpy(cmdLine, "");
      }
      else {
         // Remove whatever follows '#' character
         if (strpbrk(cmdLine, "#")) {
            *(strpbrk(cmdLine, "#")) = '\0';
         }
      }

      cmdLen = strlen(cmdLine);

      if (strchr(cmdLine, ';') != NULL) {
         curCmd = strtok(cmdLine, ";");
      }
      else {
         curCmd = cmdLine;
      }
      while (curCmd != NULL && !quit) {
         strcpy(curCommand, curCmd);
         parsedLen += strlen(curCommand);
         quit = executeCommand(curCommand, currentFile, echo, &graph,
                               sccs, scclen, ccs, cclen, params.force);
         /* Cannot use strtok command in the form
            strtok(NULL, char*), because the executeCommand
            function altered its internal buffer */
         if (parsedLen < cmdLen) {
            curCmd = strtok(curCmd + strlen(curCmd) + 1, ";");
         }
         else {
            curCmd = NULL;
         }
      }
   } while (!quit);

   clearMem(&graph, sccs, scclen, ccs, cclen);
}

/* This function parses command line parameters and
   returns them inside a parameter structure; if any
   syntax error occurs, returned value is false, and
   a short explanation is reported inside `error';
   otherwise, returned value is true */
bool parseArguments(int argc, char **argv,
                    cmdLineParameters &params,
                    char *error) {
   bool parsed, expectScriptFile;

   // Set default parameters
   params.force = false;
   params.help = false;

   strcpy(error, "");
   expectScriptFile = false;

   for (int i = 1; i < argc; i++) {
      parsed = false;
      if (expectScriptFile) {
         if (params.scriptFile != NULL) {
            strcat (error, "There is more than one script file");
            return false;
         }
         params.scriptFile = new char[strlen(argv[i]) + 1];
         strcpy(params.scriptFile, argv[i]);
         parsed = true;
         expectScriptFile = false;
      }
      if (!parsed && strcmp(argv[i], "--help") == 0) {
         parsed = true;
         params.help = true;
      }
      if (!parsed && strcmp(argv[i], "-f") == 0) {
         parsed = true;
         if (params.force) {
            strcat (error, "Option -f cannot be used more than once");
            return false;
         }
         params.force = true;
      }
      if (!parsed && strcmp(argv[i], "-s") == 0) {
         parsed = true;
         expectScriptFile = true;
      }
      if (!parsed) {
         // This argument does not appear to be a valid option
         if (params.scriptFile != NULL) {
            // Sorry, no other chance to parse it
            strcat (error, "Invalid option: `");
            strcat (error, argv[i]);
            strcat (error, "'");
            return false;
         }
         else {
            // This can be a script file name
            parsed = true;
            params.scriptFile = new char[strlen(argv[i]) + 1];
            strcpy(params.scriptFile, argv[i]);
         }
      }
   }

   if (expectScriptFile) {
      strcat (error, "Script file name expected after -s option");
      return false;
   }
   return true;
}

// This function prints out program name
void printHeader(FILE *file) {
   static bool alreadyPrinted = false;
   if (!alreadyPrinted) {
      if (isTerm(file)) {
         fprintf (file, "%stviewGraph%s 1.3\n", CONSOLE_BOLD, CONSOLE_NORMAL);
      }
      else {
         fprintf (file, "tviewGraph 1.3\n");
      }
      alreadyPrinted = true;
   }
}

// This function prints a command line usage explanation
void printHelp(FILE *file, char *cmdLine) {
   char CON_BOLD[strlen(CONSOLE_BOLD) + 1];
   char CON_UNDERLINE[strlen(CONSOLE_UNDERLINE) + 1];
   char CON_NORMAL[strlen(CONSOLE_NORMAL) + 1];
   if (isTerm(file)) {
      strcpy (CON_BOLD, CONSOLE_BOLD);
      strcpy (CON_UNDERLINE, CONSOLE_UNDERLINE);
      strcpy (CON_NORMAL, CONSOLE_NORMAL);
   }
   else {
      strcpy (CON_BOLD, "");
      strcpy (CON_UNDERLINE, "");
      strcpy (CON_NORMAL, "");
   }
   printHeader(file);
   fprintf (file, "%sTORQUE toolkit%s graph analysis tool\n", CON_BOLD, CON_NORMAL);
   fprintf (file, "USAGE: %s [options] [[-s] <scriptFile>]\n\n",
      cmdLine);
   fprintf (file, "%sOptions%s can be:\n", CON_BOLD, CON_NORMAL);
   fprintf (file, "  -f       Do not ask user (default behaviour) when overwriting an existing\n");
   fprintf (file, "           file\n");
   fprintf (file, "  --help   Show this help\n\n\n");
   fprintf (file, "<scriptFile> is the name of a file storing a sequence of tviewGraph commands;\n");
   fprintf (file, "omitting it will result in either entering an interactive mode or commands\n");
   fprintf (file, "being taken from standard input (if it has been redirected). More information\n");
   fprintf (file, "about available scripting commands can be obtained by entering interactive\n");
   fprintf (file, "mode (just use no arguments to achieve this). Use -s option when ambiguities\n");
   fprintf (file, "may arise between file name and other command line options.\n");
}

/* This function reads a script file and runs
   its commands */
void runScript(FILE *file, const cmdLineParameters &params) {
   char cmdLine[MAX_COMMAND_LENGTH + 1],
        curCommand[MAX_COMMAND_LENGTH + 1], *curCmd;
   bool echo = false, quit;
   ASGraph *graph = NULL;
   ConnCompArray ccs, sccs;
   long int scclen, cclen;
   char currentFile[513];     // One more byte for `\0'
   int parsedLen, cmdLen;

   ccs = sccs = NULL;
   scclen = cclen = 0;

   while (!feof(file) && !ferror(file) && !quit) {
      if (fgets(cmdLine, MAX_COMMAND_LENGTH, file) != NULL) {
         parsedLen = 0;
         quit = false;

         // Remove CR/LF at end of line
         if (strpbrk(cmdLine, "\n\r")) {
            *(strpbrk(cmdLine, "\n\r")) = '\0';
         }

         // Remove comments
         if (cmdLine[0] == '#') {
            // The whole line is a comment
            strcpy(cmdLine, "");
         }
         else {
            // Remove whatever follows '#' character
            if (strpbrk(cmdLine, "#")) {
               *(strpbrk(cmdLine, "#")) = '\0';
            }
         }

         cmdLen = strlen(cmdLine);

         if (strchr(cmdLine, ';') != NULL) {
            curCmd = strtok(cmdLine, ";");
         }
         else {
            curCmd = cmdLine;
         }
         while (curCmd != NULL && !quit) {
            strcpy(curCommand, curCmd);
            parsedLen += strlen(curCommand);
            quit = executeCommand(curCommand, currentFile, echo, &graph,
                                  sccs, scclen, ccs, cclen, params.force);
            /* Cannot use strtok command in the form
               strtok(NULL, char*), because the executeCommand
               function altered its internal buffer */
            if (parsedLen < cmdLen) {
               curCmd = strtok(curCmd + strlen(curCmd) + 1, ";");
            }
            else {
               curCmd = NULL;
            }
         }
      }
   }

   clearMem(&graph, sccs, scclen, ccs, cclen);
}

int main(int argc, char **argv) {
   cmdLineParameters params;
   FILE *scriptFile;
   char error[513];              // One more byte for '\0'
   bool stdinIsTerm, stdoutIsTerm;

   if (!parseArguments(argc, argv, params, error)) {
      printHeader (stderr);
      fprintf (stderr, "Syntax error: %s\n", error);
      fprintf (stderr, "Type %s --help for more information\n", argv[0]);
      return 1;
   }

   // Check which streams are connected to a terminal
   stdinIsTerm = isTerm(stdin);
   stdoutIsTerm = isTerm(stdout);

   if (params.help) {
      printHelp (stdout, argv[0]);
      return 0;
   }

   if (stdinIsTerm && stdoutIsTerm) {
      printHeader (stdout);
   }

   if (params.scriptFile != NULL) {
      scriptFile = fopen("file", "r");
      if (scriptFile == NULL) {
         fprintf (stderr, "Unable to open file `%s' for reading\n", params.scriptFile);
         fprintf (stderr, "(%s)\n", strerror(errno));
         return 2;
      }

      // Run script commands
      runScript(scriptFile, params);
      fclose(scriptFile);
   }
   else {
      if (!isTerm(stdin)) {
         // Assume commands come from stdin
         runScript(stdin, params);
      }
      else {
         goInteractive(params);
      }
   }
}
